# cython: language_level=3
# cython: profile=True
# Time-stamp: <2024-10-22 16:59:53 Tao Liu>

"""Module for SAPPER PosReadsInfo class.

Copyright (c) 2017 Tao Liu <tliu4@buffalo.edu>

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).

@status:  experimental
@version: $Revision$
@author:  Tao Liu
@contact: tliu4@buffalo.edu
"""

# ------------------------------------
# python modules
# ------------------------------------
from MACS3.Signal.VariantStat import (CalModel_Homo,
                                      CalModel_Heter_noAS,
                                      CalModel_Heter_AS)
# calculate_GQ,
# calculate_GQ_heterASsig)
from MACS3.Signal.Prob import binomial_cdf
from MACS3.Signal.PeakVariants import Variant

import cython
import numpy as np
import cython.cimports.numpy as cnp
from cython.cimports.cpython import bool

LN10 = 2.3025850929940458

# ------------------------------------
# constants
# ------------------------------------
__version__ = "Parser $Revision$"
__author__ = "Tao Liu <tliu4@buffalo.edu>"
__doc__ = "All Parser classes"

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Classes
# ------------------------------------


@cython.cclass
class PosReadsInfo:
    ref_pos: cython.long
    ref_allele: cython.bytes
    alt_allele: cython.bytes
    filterout: bool          # if true, do not output
    bq_set_T: dict       # {A:[], C:[], G:[], T:[], N:[]} for treatment
    bq_set_C: dict
    n_reads_T: dict      # {A:[], C:[], G:[], T:[], N:[]} for treatment
    n_reads_C: dict
    n_reads: dict
    n_strand: list  # [{A:[], C:[], G:[], T:[], N:[]},{A:[], C:[], G:[], T:[], N:[]}] for total appearance on plus strand and minus strand for ChIP sample only
    n_tips: dict  # count of nt appearing at tips
    top1allele: cython.bytes
    top2allele: cython.bytes
    top12alleles_ratio: cython.float
    lnL_homo_major: cython.double
    lnL_heter_AS: cython.double
    lnL_heter_noAS: cython.double
    lnL_homo_minor: cython.double
    BIC_homo_major: cython.double
    BIC_heter_AS: cython.double
    BIC_heter_noAS: cython.double
    BIC_homo_minor: cython.double
    PL_00: cython.double
    PL_01: cython.double
    PL_11: cython.double
    deltaBIC: cython.double
    heter_noAS_kc: cython.int
    heter_noAS_ki: cython.int
    heter_AS_kc: cython.int
    heter_AS_ki: cython.int
    heter_AS_alleleratio: cython.double
    GQ_homo_major: cython.int
    GQ_heter_noAS: cython.int
    GQ_heter_AS: cython.int  # phred scale of prob by standard formular
    GQ_heter_ASsig: cython.int  # phred scale of prob, to measure the difference between AS and noAS
    GQ: cython.double
    GT: str
    type: str
    mutation_type: str       # SNV or Insertion or Deletion
    hasfermiinfor: bool  # if no fermi bam overlap in the position, false; if fermi bam in the position GT: N, false; if anyone of top2allele is not in fermi GT NTs, false;
    fermiNTs: bytearray

    def __cinit__(self):
        self.filterout = False
        self.GQ = 0
        self.GT = "unsure"
        self.alt_allele = b'.'

    def __init__(self,
                 ref_pos: cython.long,
                 ref_allele: cython.bytes):
        self.ref_pos = ref_pos
        self.ref_allele = ref_allele
        self.bq_set_T = {ref_allele: [], b'A': [], b'C': [], b'G': [], b'T': [], b'N': [], b'*': []}
        self.bq_set_C = {ref_allele: [], b'A': [], b'C': [], b'G': [], b'T': [], b'N': [], b'*': []}
        self.n_reads_T = {ref_allele: 0, b'A': 0, b'C': 0, b'G': 0, b'T': 0, b'N': 0, b'*': 0}
        self.n_reads_C = {ref_allele: 0, b'A': 0, b'C': 0, b'G': 0, b'T': 0, b'N': 0, b'*': 0}
        self.n_reads = {ref_allele: 0, b'A': 0, b'C': 0, b'G': 0, b'T': 0, b'N': 0, b'*': 0}
        self.n_strand = [{ref_allele: 0, b'A': 0, b'C': 0, b'G': 0, b'T': 0, b'N': 0, b'*': 0},
                         {ref_allele: 0, b'A': 0, b'C': 0, b'G': 0, b'T': 0, b'N': 0, b'*': 0}]
        self.n_tips = {ref_allele: 0, b'A': 0, b'C': 0, b'G': 0, b'T': 0, b'N': 0, b'*': 0}

    def __getstate__(self):
        return (self.ref_pos,
                self.ref_allele,
                self.alt_allele,
                self.filterout,
                self.bq_set_T,
                self.bq_set_C,
                self.n_reads_T,
                self.n_reads_C,
                self.n_reads,
                self.n_strand,
                self.n_tips,
                self.top1allele,
                self.top2allele,
                self.top12alleles_ratio,
                self.lnL_homo_major,
                self.lnL_heter_AS,
                self.lnL_heter_noAS,
                self.lnL_homo_minor,
                self.BIC_homo_major,
                self.BIC_heter_AS,
                self.BIC_heter_noAS,
                self.BIC_homo_minor,
                self.heter_noAS_kc,
                self.heter_noAS_ki,
                self.heter_AS_kc,
                self.heter_AS_ki,
                self.heter_AS_alleleratio,
                self.GQ_homo_major,
                self.GQ_heter_noAS,
                self.GQ_heter_AS,
                self.GQ_heter_ASsig,
                self.GQ,
                self.GT,
                self.type,
                self.hasfermiinfor,
                self.fermiNTs)

    def __setstate__(self, state):
        (self.ref_pos,
         self.ref_allele,
         self.alt_allele,
         self.filterout,
         self.bq_set_T,
         self.bq_set_C,
         self.n_reads_T,
         self.n_reads_C,
         self.n_reads,
         self.n_strand,
         self.n_tips,
         self.top1allele,
         self.top2allele,
         self.top12alleles_ratio,
         self.lnL_homo_major,
         self.lnL_heter_AS,
         self.lnL_heter_noAS,
         self.lnL_homo_minor,
         self.BIC_homo_major,
         self.BIC_heter_AS,
         self.BIC_heter_noAS,
         self.BIC_homo_minor,
         self.heter_noAS_kc,
         self.heter_noAS_ki,
         self.heter_AS_kc,
         self.heter_AS_ki,
         self.heter_AS_alleleratio,
         self.GQ_homo_major,
         self.GQ_heter_noAS,
         self.GQ_heter_AS,
         self.GQ_heter_ASsig,
         self.GQ,
         self.GT,
         self.type,
         self.hasfermiinfor,
         self.fermiNTs) = state

    @cython.ccall
    def filterflag(self) -> bool:
        return self.filterout

    @cython.ccall
    def apply_GQ_cutoff(self,
                        min_homo_GQ: cython.int = 50,
                        min_heter_GQ: cython.int = 100):
        if self.filterout:
            return
        if self.type.startswith('homo') and self.GQ < min_homo_GQ:
            self.filterout = True
        elif self.type.startswith('heter') and self.GQ < min_heter_GQ:
            self.filterout = True
        return

    @cython.ccall
    def apply_deltaBIC_cutoff(self,
                              min_delta_BIC: cython.float = 10):
        if self.filterout:
            return
        if self.deltaBIC < min_delta_BIC:
            self.filterout = True
        return

    @cython.ccall
    def add_T(self,
              read_index: cython.int,
              read_allele: cython.bytes,
              read_bq: cython.int,
              strand: cython.int,
              tip: bool,
              Q: cython.int = 20):
        """ Strand 0: plus, 1: minus

        Q is the quality cutoff. By default, only consider Q20 or read_bq > 20.
        """
        if read_bq <= Q:
            return
        if not self.n_reads.has_key(read_allele):
            self.bq_set_T[read_allele] = []
            self.bq_set_C[read_allele] = []
            self.n_reads_T[read_allele] = 0
            self.n_reads_C[read_allele] = 0
            self.n_reads[read_allele] = 0
            self.n_strand[0][read_allele] = 0
            self.n_strand[1][read_allele] = 0
            self.n_tips[read_allele] = 0
        self.bq_set_T[read_allele].append(read_bq)
        self.n_reads_T[read_allele] += 1
        self.n_reads[read_allele] += 1
        self.n_strand[strand][read_allele] += 1
        if tip:
            self.n_tips[read_allele] += 1

    @cython.ccall
    def add_C(self,
              read_index: cython.int,
              read_allele: cython.bytes,
              read_bq: cython.int,
              strand: cython.int,
              Q: cython.int = 20):
        if read_bq <= Q:
            return
        if not self.n_reads.has_key(read_allele):
            self.bq_set_T[read_allele] = []
            self.bq_set_C[read_allele] = []
            self.n_reads_T[read_allele] = 0
            self.n_reads_C[read_allele] = 0
            self.n_reads[read_allele] = 0
            self.n_strand[0][read_allele] = 0
            self.n_strand[1][read_allele] = 0
            self.n_tips[read_allele] = 0
        self.bq_set_C[read_allele].append(read_bq)
        self.n_reads_C[read_allele] += 1
        self.n_reads[read_allele] += 1

    @cython.ccall
    def raw_read_depth(self,
                       opt: str = "all") -> cython.int:
        if opt == "all":
            return sum(self.n_reads.values())
        elif opt == "T":
            return sum(self.n_reads_T.values())
        elif opt == "C":
            return sum(self.n_reads_C.values())
        else:
            raise Exception("opt should be either 'all', 'T' or 'C'.")

    @cython.ccall
    def update_top_alleles(self,
                           min_top12alleles_ratio: cython.float = 0.8,
                           min_altallele_count: cython.int = 2,
                           max_allowed_ar: cython.float = 0.95):
        """Identify top1 and top2 NT.  the ratio of (top1+top2)/total
        """
        [self.top1allele, self.top2allele] = sorted(self.n_reads,
                                                    key=self.n_reads_T.get,
                                                    reverse=True)[:2]

        # if top2 allele count in ChIP is lower than
        # min_altallele_count, or when allele ratio top1/(top1+top2)
        # is larger than max_allowed_ar in ChIP, we won't consider
        # this allele at all.  we set values of top2 allele in
        # dictionaries to [] and ignore top2 allele entirely.
        # max(self.n_strand[0][self.top2allele], self.n_strand[1][self.top2allele]) < min_altallele_count
        # if self.ref_pos == 52608504:
        #    prself: cython.int.ref_pos, self.n_reads_T[self.top1allele], self.n_reads_T[self.top2allele], self.n_reads_C[self.top1allele], self.n_reads_C[self.top2allele]
        if self.n_reads_T[self.top1allele] + self.n_reads_T[self.top2allele] == 0:
            self.filterout = True
            return

        if (len(self.top1allele) == 1 and len(self.top2allele) == 1) and (self.top2allele != self.ref_allele and ((self.n_reads_T[self.top2allele] - self.n_tips[self.top2allele]) < min_altallele_count) or self.n_reads_T[self.top1allele]/(self.n_reads_T[self.top1allele] + self.n_reads_T[self.top2allele]) > max_allowed_ar):
            self.bq_set_T[self.top2allele] = []
            self.bq_set_C[self.top2allele] = []
            self.n_reads_T[self.top2allele] = 0
            self.n_reads_C[self.top2allele] = 0
            self.n_reads[self.top2allele] = 0
            self.n_tips[self.top2allele] = 0
            if (self.top1allele != self.ref_allele and (self.n_reads_T[self.top1allele] - self.n_tips[self.top1allele]) < min_altallele_count):
                self.bq_set_T[self.top1allele] = []
                self.bq_set_C[self.top1allele] = []
                self.n_reads_T[self.top1allele] = 0
                self.n_reads_C[self.top1allele] = 0
                self.n_reads[self.top1allele] = 0
                self.n_tips[self.top1allele] = 0

        if self.n_reads_T[self.top1allele] + self.n_reads_T[self.top2allele] == 0:
            self.filterout = True
            return

        self.top12alleles_ratio = (self.n_reads[self.top1allele] + self.n_reads[self.top2allele]) / sum(self.n_reads.values())
        if self.top12alleles_ratio < min_top12alleles_ratio:
            self.filterout = True
            return

        if self.top1allele == self.ref_allele and self.n_reads[self.top2allele] == 0:
            # This means this position only contains top1allele which is the ref_allele. So the GT must be 0/0
            self.type = "homo_ref"
            self.filterout = True
            return
        return

    @cython.ccall
    def top12alleles(self):
        print(self.ref_pos, self.ref_allele)
        print("Top1allele", self.top1allele, "Treatment",
              self.bq_set_T[self.top1allele], "Control",
              self.bq_set_C[self.top1allele])
        print("Top2allele", self.top2allele, "Treatment",
              self.bq_set_T[self.top2allele], "Control",
              self.bq_set_C[self.top2allele])
    
    @cython.ccall
    def call_GT(self, max_allowed_ar: cython.float = 0.99):
        """Require update_top_alleles being called.
        """
        top1_bq_T: cnp.ndarray(cython.int, ndim=1)
        top2_bq_T: cnp.ndarray(cython.int, ndim=1)
        top1_bq_C: cnp.ndarray(cython.int, ndim=1)
        top2_bq_C: cnp.ndarray(cython.int, ndim=1)
        tmp_mutation_type: list
        tmp_alt: cython.bytes

        if self.filterout:
            return

        top1_bq_T = np.array(self.bq_set_T[self.top1allele], dtype="i4")
        top2_bq_T = np.array(self.bq_set_T[self.top2allele], dtype="i4")
        top1_bq_C = np.array(self.bq_set_C[self.top1allele], dtype="i4")
        top2_bq_C = np.array(self.bq_set_C[self.top2allele], dtype="i4")
        (self.lnL_homo_major, self.BIC_homo_major) = CalModel_Homo(top1_bq_T,
                                                                   top1_bq_C,
                                                                   top2_bq_T,
                                                                   top2_bq_C)
        (self.lnL_homo_minor, self.BIC_homo_minor) = CalModel_Homo(top2_bq_T,
                                                                   top2_bq_C,
                                                                   top1_bq_T,
                                                                   top1_bq_C)
        (self.lnL_heter_noAS, self.BIC_heter_noAS) = CalModel_Heter_noAS(top1_bq_T,
                                                                         top1_bq_C,
                                                                         top2_bq_T,
                                                                         top2_bq_C)
        (self.lnL_heter_AS, self.BIC_heter_AS) = CalModel_Heter_AS(top1_bq_T,
                                                                   top1_bq_C,
                                                                   top2_bq_T,
                                                                   top2_bq_C,
                                                                   max_allowed_ar)

        # if self.ref_pos == 71078525:
        #    print "---"
        #    prlen: cython.int(top1_bq_T), len(top1_bq_C), len(top2_bq_T), len(top2_bq_C)
        #    prself: cython.int.lnL_homo_major, self.lnL_homo_minor, self.lnL_heter_noAS, self.lnL_heter_AS
        #    prself: cython.int.BIC_homo_major, self.BIC_homo_minor, self.BIC_heter_noAS, self.BIC_heter_AS

        if self.top1allele != self.ref_allele and self.n_reads[self.top2allele] == 0:
            # in this case, there is no top2 nt (or socalled minor
            # allele) in either treatment or control, we should assume
            # it's a 1/1 genotype. We will take 1/1 if it passes BIC
            # test (deltaBIC >=2), and will skip this loci if it can't
            # pass the test.

            self.deltaBIC = min(self.BIC_heter_noAS, self.BIC_heter_AS, self.BIC_homo_minor) - self.BIC_homo_major
            if self.deltaBIC < 2:
                self.filterout = True
                return

            self.type = "homo"
            self.GT = "1/1"

            self.PL_00 = -10.0 * self.lnL_homo_minor / LN10
            self.PL_01 = -10.0 * max(self.lnL_heter_noAS, self.lnL_heter_AS) / LN10
            self.PL_11 = -10.0 * self.lnL_homo_major / LN10

            self.PL_00 = max(0, self.PL_00 - self.PL_11)
            self.PL_01 = max(0, self.PL_01 - self.PL_11)
            self.PL_11 = 0

            self.GQ = min(self.PL_00, self.PL_01)
            self.alt_allele = self.top1allele
        else:
            # assign GQ, GT, and type
            if self.ref_allele != self.top1allele and self.BIC_homo_major + 2 <= self.BIC_homo_minor and self.BIC_homo_major + 2 <= self.BIC_heter_noAS and self.BIC_homo_major + 2 <= self.BIC_heter_AS:
                self.type = "homo"
                self.deltaBIC = min(self.BIC_heter_noAS, self.BIC_heter_AS, self.BIC_homo_minor) - self.BIC_homo_major
                self.GT = "1/1"
                self.alt_allele = self.top1allele

                self.PL_00 = -10.0 * self.lnL_homo_minor / LN10
                self.PL_01 = -10.0 * max(self.lnL_heter_noAS, self.lnL_heter_AS) / LN10
                self.PL_11 = -10.0 * self.lnL_homo_major / LN10

                self.PL_00 = self.PL_00 - self.PL_11
                self.PL_01 = self.PL_01 - self.PL_11
                self.PL_11 = 0

                self.GQ = min(self.PL_00, self.PL_01)

            elif self.BIC_heter_noAS + 2 <= self.BIC_homo_major and self.BIC_heter_noAS + 2 <= self.BIC_homo_minor and self.BIC_heter_noAS + 2 <= self.BIC_heter_AS:
                self.type = "heter_noAS"
                self.deltaBIC = min(self.BIC_homo_major, self.BIC_homo_minor) - self.BIC_heter_noAS

                self.PL_00 = -10.0 * self.lnL_homo_minor / LN10
                self.PL_01 = -10.0 * self.lnL_heter_noAS / LN10
                self.PL_11 = -10.0 * self.lnL_homo_major / LN10

                self.PL_00 = self.PL_00 - self.PL_01
                self.PL_11 = self.PL_11 - self.PL_01
                self.PL_01 = 0

                self.GQ = min(self.PL_00, self.PL_11)

            elif self.BIC_heter_AS + 2 <= self.BIC_homo_major and self.BIC_heter_AS + 2 <= self.BIC_homo_minor and self.BIC_heter_AS + 2 <= self.BIC_heter_noAS:
                self.type = "heter_AS"
                self.deltaBIC = min(self.BIC_homo_major, self.BIC_homo_minor) - self.BIC_heter_AS

                self.PL_00 = -10.0 * self.lnL_homo_minor / LN10
                self.PL_01 = -10.0 * self.lnL_heter_AS / LN10
                self.PL_11 = -10.0 * self.lnL_homo_major / LN10

                self.PL_00 = self.PL_00 - self.PL_01
                self.PL_11 = self.PL_11 - self.PL_01
                self.PL_01 = 0

                self.GQ = min(self.PL_00, self.PL_11)

            elif self.BIC_heter_AS + 2 <= self.BIC_homo_major and self.BIC_heter_AS + 2 <= self.BIC_homo_minor:
                # can't decide if it's noAS or AS
                self.type = "heter_unsure"
                self.deltaBIC = min(self.BIC_homo_major, self.BIC_homo_minor) - max(self.BIC_heter_AS, self.BIC_heter_noAS)

                self.PL_00 = -10.0 * self.lnL_homo_minor / LN10
                self.PL_01 = -10.0 * max(self.lnL_heter_noAS, self.lnL_heter_AS) / LN10
                self.PL_11 = -10.0 * self.lnL_homo_major / LN10

                self.PL_00 = self.PL_00 - self.PL_01
                self.PL_11 = self.PL_11 - self.PL_01
                self.PL_01 = 0

                self.GQ = min(self.PL_00, self.PL_11)

            elif self.ref_allele == self.top1allele and self.BIC_homo_major < self.BIC_homo_minor and self.BIC_homo_major < self.BIC_heter_noAS and self.BIC_homo_major < self.BIC_heter_AS:
                self.type = "homo_ref"
                # we do not calculate GQ if type is homo_ref
                self.GT = "0/0"
                self.filterout = True
            else:
                self.type = "unsure"
                self.filterout = True

            if self.type.startswith("heter"):
                if self.ref_allele == self.top1allele:
                    self.alt_allele = self.top2allele
                    self.GT = "0/1"
                elif self.ref_allele == self.top2allele:
                    self.alt_allele = self.top1allele
                    self.GT = "0/1"
                else:
                    self.alt_allele = self.top1allele+b','+self.top2allele
                    self.GT = "1/2"

        tmp_mutation_type = []
        for tmp_alt in self.alt_allele.split(b','):
            if tmp_alt == b'*':
                tmp_mutation_type.append("Deletion")
            elif len(tmp_alt) > 1:
                tmp_mutation_type.append("Insertion")
            else:
                tmp_mutation_type.append("SNV")
        self.mutation_type = ",".join(tmp_mutation_type)
        return

    @cython.cfunc
    def SB_score_ChIP(self,
                      a: cython.int,
                      b: cython.int,
                      c: cython.int,
                      d: cython.int) -> cython.float:
        """ calculate score for filtering variants with strange strand biases.

        a: top1/major allele plus strand
        b: top2/minor allele plus strand
        c: top1/major allele minus strand
        d: top2/minor allele minus strand

        Return a value: cython.float so that if this value >= 1, the variant will be filtered out.
        """
        p1_l: cython.double
        p1_r: cython.double
        p2_l: cython.double
        p2_r: cython.double

        if a + b == 0 or c + d == 0:
            # if major allele and minor allele both bias to the same strand, allow it
            return 0.0

        # Rule:
        # if there is bias in top2 allele then bias in top1 allele should not be significantly smaller than it.
        # or there is no significant bias (0.5) in top2 allele.

        # pra: cython.int, b, c, d
        p1_l = binomial_cdf(a, (a+c), 0.5, lower=True)      # alternative: less than 0.5
        p1_r = binomial_cdf(c, (a+c), 0.5, lower=True)                   # greater than 0.5
        p2_l = binomial_cdf(b, (b+d), 0.5, lower=True)      # alternative: less than 0.5
        p2_r = binomial_cdf(d, (b+d), 0.5, lower=True)                   # greater than 0.5
        # prp1_l: cython.int, p1_r, p2_l, p2_r

        if (p1_l < 0.05 and p2_r < 0.05) or (p1_r < 0.05 and p2_l < 0.05):
            # we reject loci where the significant biases are inconsistent between top1 and top2 alleles.
            return 1.0
        else:
            # if b<=2 and d=0 or b=0 and d<=2 -- highly possible FPs
            # if (b<=2 and d==0 or b==0 and d<=2):
            #    return 1
            # can't decide
            return 0.0

    @cython.cfunc
    def SB_score_ATAC(self,
                      a: cython.int,
                      b: cython.int,
                      c: cython.int,
                      d: cython.int) -> cython.float:
        """ calculate score for filtering variants with strange strand biases.

        ATAC-seq version

        a: top1/major allele plus strand
        b: top2/minor allele plus strand
        c: top1/major allele minus strand
        d: top2/minor allele minus strand

        Return a value: cython.float so that if this value >= 1, the variant will be filtered out.
        """
        p1_l: cython.double
        p1_r: cython.double
        p2_l: cython.double
        p2_r: cython.double

        if a+b == 0 or c+d == 0:
            # if major allele and minor allele both bias to the same strand, allow it
            return 0.0

        # Rule:
        # if there is bias in top2 allele then bias in top1 allele should not be significantly smaller than it.
        # or there is no significant bias (0.5) in top2 allele.
        # pra: cython.int, b, c, d
        p1_l = binomial_cdf(a, (a+c), 0.5, lower=True)      # alternative: less than 0.5
        p1_r = binomial_cdf(c, (a+c), 0.5, lower=True)      #              greater than 0.5
        p2_l = binomial_cdf(b, (b+d), 0.5, lower=True)      # alternative: less than 0.5
        p2_r = binomial_cdf(d, (b+d), 0.5, lower=True)      #              greater than 0.5
        # prp1_l: cython.int, p1_r, p2_l, p2_r

        if (p1_l < 0.05 and p2_r < 0.05) or (p1_r < 0.05 and p2_l < 0.05):
            # we reject loci where the significant biases are inconsistent between top1 and top2 alleles.
            return 1.0
        else:
            # can't decide
            return 0.0

    @cython.ccall
    def to_vcf(self) -> str:
        """Output REF,ALT,QUAL,FILTER,INFO,FORMAT, SAMPLE columns.
        """
        vcf_ref: str
        vcf_alt: str
        vcf_qual: str
        vcf_filter: str
        vcf_info: str
        vcf_format: str
        vcf_sample: str

        vcf_ref = self.ref_allele.decode()
        vcf_alt = self.alt_allele.decode()
        vcf_qual = "%d" % self.GQ
        vcf_filter = "."
        vcf_info = (b"M=%s;MT=%s;DPT=%d;DPC=%d;DP1T=%d%s;DP2T=%d%s;DP1C=%d%s;DP2C=%d%s;SB=%d,%d,%d,%d;DBIC=%.2f;BICHOMOMAJOR=%.2f;BICHOMOMINOR=%.2f;BICHETERNOAS=%.2f;BICHETERAS=%.2f;AR=%.2f" %
                    (self.type.encode(),
                     self.mutation_type.encode(),
                     sum(self.n_reads_T.values()),
                     sum(self.n_reads_C.values()),
                     self.n_reads_T[self.top1allele],
                     self.top1allele,
                     self.n_reads_T[self.top2allele],
                     self.top2allele,
                     self.n_reads_C[self.top1allele],
                     self.top1allele,
                     self.n_reads_C[self.top2allele],
                     self.top2allele,
                     self.n_strand[0][self.top1allele],
                     self.n_strand[0][self.top2allele],
                     self.n_strand[1][self.top1allele],
                     self.n_strand[1][self.top2allele],
                     self.deltaBIC,
                     self.BIC_homo_major,
                     self.BIC_homo_minor,
                     self.BIC_heter_noAS,
                     self.BIC_heter_AS,
                     self.n_reads_T[self.top1allele]/(self.n_reads_T[self.top1allele]+self.n_reads_T[self.top2allele])
                    )).decode()
        vcf_format = "GT:DP:GQ:PL"
        vcf_sample = "%s:%d:%d:%d,%d,%d" % (self.GT, self.raw_read_depth(opt="all"), self.GQ, self.PL_00, self.PL_01, self.PL_11)
        return "\t".join((vcf_ref, vcf_alt, vcf_qual, vcf_filter, vcf_info, vcf_format, vcf_sample))

    @cython.ccall
    def toVariant(self):
        v: Variant

        v = Variant(self.ref_allele.decode(),
                    self.alt_allele.decode(),
                    self.GQ,
                    '.',
                    self.type,
                    self.mutation_type,
                    self.top1allele.decode(),
                    self.top2allele.decode(),
                    sum(self.n_reads_T.values()),
                    sum(self.n_reads_C.values()),
                    self.n_reads_T[self.top1allele],
                    self.n_reads_T[self.top2allele],
                    self.n_reads_C[self.top1allele],
                    self.n_reads_C[self.top2allele],
                    self.n_strand[0][self.top1allele],
                    self.n_strand[0][self.top2allele],
                    self.n_strand[1][self.top1allele],
                    self.n_strand[1][self.top2allele],
                    self.deltaBIC,
                    self.BIC_homo_major,
                    self.BIC_homo_minor,
                    self.BIC_heter_noAS,
                    self.BIC_heter_AS,
                    self.n_reads_T[self.top1allele]/(self.n_reads_T[self.top1allele]+self.n_reads_T[self.top2allele]),
                    self.GT,
                    self.raw_read_depth(opt="all"),
                    self.PL_00,
                    self.PL_01,
                    self.PL_11)
        return v



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2024-10-22 17:12:29 Tao Liu>

"""Module for SAPPER PeakVariants class.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
from copy import copy
import cython
from cython.cimports.cpython import bool


@cython.cclass
class Variant:
    v_ref_pos: cython.long
    v_ref_allele: str
    v_alt_allele: str
    v_GQ: cython.int
    v_filter: str
    v_type: str
    v_mutation_type: str
    v_top1allele: str
    v_top2allele: str
    v_DPT: cython.int
    v_DPC: cython.int
    v_DP1T: cython.int
    v_DP2T: cython.int
    v_DP1C: cython.int
    v_DP2C: cython.int
    v_PLUS1T: cython.int
    v_PLUS2T: cython.int
    v_MINUS1T: cython.int
    v_MINUS2T: cython.int
    v_deltaBIC: cython.float
    v_BIC_homo_major: cython.float
    v_BIC_homo_minor: cython.float
    v_BIC_heter_noAS: cython.float
    v_BIC_heter_AS: cython.float
    v_AR: cython.float
    v_GT: str
    v_DP: cython.int
    v_PL_00: cython.int
    v_PL_01: cython.int
    v_PL_11: cython.int

    def __init__(self,
                 ref_allele: str,
                 alt_allele: str,
                 GQ: cython.int,
                 filter: str,
                 type: str,
                 mutation_type: str,
                 top1allele: str,
                 top2allele: str,
                 DPT: cython.int,
                 DPC: cython.int,
                 DP1T: cython.int,
                 DP2T: cython.int,
                 DP1C: cython.int,
                 DP2C: cython.int,
                 PLUS1T: cython.int,
                 PLUS2T: cython.int,
                 MINUS1T: cython.int,
                 MINUS2T: cython.int,
                 deltaBIC: cython.float,
                 BIC_homo_major: cython.float,
                 BIC_homo_minor: cython.float,
                 BIC_heter_noAS: cython.float,
                 BIC_heter_AS: cython.float,
                 AR: cython.float,
                 GT: str,
                 DP: cython.int,
                 PL_00: cython.int,
                 PL_01: cython.int,
                 PL_11: cython.int):
        self.v_ref_allele = ref_allele
        self.v_alt_allele = alt_allele
        self.v_GQ = GQ
        self.v_filter = filter
        self.v_type = type
        self.v_mutation_type = mutation_type
        self.v_top1allele = top1allele
        self.v_top2allele = top2allele
        self.v_DPT = DPT
        self.v_DPC = DPC
        self.v_DP1T = DP1T
        self.v_DP2T = DP2T
        self.v_DP1C = DP1C
        self.v_DP2C = DP2C
        self.v_PLUS1T = PLUS1T
        self.v_PLUS2T = PLUS2T
        self.v_MINUS1T = MINUS1T
        self.v_MINUS2T = MINUS2T
        self.v_deltaBIC = deltaBIC
        self.v_BIC_homo_major = BIC_homo_major
        self.v_BIC_homo_minor = BIC_homo_minor
        self.v_BIC_heter_noAS = BIC_heter_noAS
        self.v_BIC_heter_AS = BIC_heter_AS
        self.v_AR = AR
        self.v_GT = GT
        self.v_DP = DP
        self.v_PL_00 = PL_00
        self.v_PL_01 = PL_01
        self.v_PL_11 = PL_11

    def __getstate__(self):
        # self.v_ref_pos,
        return (self.v_ref_allele,
                self.v_alt_allele,
                self.v_GQ,
                self.v_filter,
                self.v_type,
                self.v_mutation_type,
                self.v_top1allele,
                self.v_top2allele,
                self.v_DPT,
                self.v_DPC,
                self.v_DP1T,
                self.v_DP2T,
                self.v_DP1C,
                self.v_DP2C,
                self.v_PLUS1T,
                self.v_PLUS2T,
                self.v_MINUS1T,
                self.v_MINUS2T,
                self.v_deltaBIC,
                self.v_BIC_homo_major,
                self.v_BIC_homo_minor,
                self.v_BIC_heter_noAS,
                self.v_BIC_heter_AS,
                self.v_AR,
                self.v_GT,
                self.v_DP,
                self.v_PL_00,
                self.v_PL_01,
                self.v_PL_11)

    def __setstate__(self, state):
        # self.v_ref_pos,
        (self.v_ref_allele,
         self.v_alt_allele,
         self.v_GQ,
         self.v_filter,
         self.v_type,
         self.v_mutation_type,
         self.v_top1allele,
         self.v_top2allele,
         self.v_DPT,
         self.v_DPC,
         self.v_DP1T,
         self.v_DP2T,
         self.v_DP1C,
         self.v_DP2C,
         self.v_PLUS1T,
         self.v_PLUS2T,
         self.v_MINUS1T,
         self.v_MINUS2T,
         self.v_deltaBIC,
         self.v_BIC_homo_major,
         self.v_BIC_homo_minor,
         self.v_BIC_heter_noAS,
         self.v_BIC_heter_AS,
         self.v_AR,
         self.v_GT,
         self.v_DP,
         self.v_PL_00,
         self.v_PL_01,
         self.v_PL_11) = state

    @cython.ccall
    def is_indel(self) -> bool:
        if self.v_mutation_type.find("Insertion") != -1 or self.v_mutation_type.find("Deletion") != -1:
            return True
        else:
            return False

    @cython.ccall
    def is_only_del(self) -> bool:
        if self.v_mutation_type == "Deletion":
            return True
        else:
            return False

    @cython.ccall
    def is_only_insertion(self) -> bool:
        if self.v_mutation_type == "Insertion":
            return True
        else:
            return False

    def __getitem__(self, keyname):
        if keyname == "ref_allele":
            return self.v_ref_allele
        elif keyname == "alt_allele":
            return self.v_alt_allele
        elif keyname == "top1allele":
            return self.v_top1allele
        elif keyname == "top2allele":
            return self.v_top2allele
        elif keyname == "type":
            return self.type
        elif keyname == "mutation_type":
            return self.mutation_type
        else:
            raise Exception("keyname is not accessible:", keyname)

    def __setitem__(self, keyname, v):
        if keyname == "ref_allele":
            self.v_ref_allele = v
        elif keyname == "alt_allele":
            self.v_alt_allele = v
        elif keyname == "top1allele":
            self.v_top1allele = v
        elif keyname == "top2allele":
            self.v_top2allele = v
        elif keyname == "type":
            self.type = v
        elif keyname == "mutation_type":
            self.mutation_type = v
        else:
            raise Exception("keyname is not accessible:", keyname)

    @cython.ccall
    def is_refer_biased_01(self,
                           ar: cython.float = 0.85) -> bool:
        if self.v_AR >= ar and self.v_ref_allele == self.v_top1allele:
            return True
        else:
            return False

    @cython.ccall
    def top1isreference(self) -> bool:
        if self.v_ref_allele == self.v_top1allele:
            return True
        else:
            return False

    @cython.ccall
    def top2isreference(self) -> bool:
        if self.v_ref_allele == self.v_top2allele:
            return True
        else:
            return False        
        
    @cython.ccall
    def toVCF(self) -> str:
        return "\t".join((self.v_ref_allele, self.v_alt_allele, "%d" % self.v_GQ, self.v_filter,
                          "M=%s;MT=%s;DPT=%d;DPC=%d;DP1T=%d%s;DP2T=%d%s;DP1C=%d%s;DP2C=%d%s;SB=%d,%d,%d,%d;DBIC=%.2f;BICHOMOMAJOR=%.2f;BICHOMOMINOR=%.2f;BICHETERNOAS=%.2f;BICHETERAS=%.2f;AR=%.2f" %
                          (self.v_type, self.v_mutation_type, self.v_DPT, self.v_DPC, self.v_DP1T, self.v_top1allele,
                           self.v_DP2T, self.v_top2allele, self.v_DP1C, self.v_top1allele, self.v_DP2C, self.v_top2allele,
                           self.v_PLUS1T, self.v_PLUS2T, self.v_MINUS1T, self.v_MINUS2T,
                           self.v_deltaBIC,
                           self.v_BIC_homo_major, self.v_BIC_homo_minor, self.v_BIC_heter_noAS,self.v_BIC_heter_AS,
                           self.v_AR
                           ),
                          "GT:DP:GQ:PL",
                          "%s:%d:%d:%d,%d,%d" % (self.v_GT, self.v_DP, self.v_GQ, self.v_PL_00, self.v_PL_01, self.v_PL_11)
                          ))


@cython.cclass
class PeakVariants:
    chrom: str
    d_Variants: dict
    start: cython.long
    end: cython.long
    refseq: bytes

    def __init__(self,
                 chrom: str,
                 start: cython.long,
                 end: cython.long,
                 s: bytes):
        self.chrom = chrom
        self.d_Variants = {}
        self.start = start
        self.end = end
        self.refseq = s

    def __getstate__(self):
        return (self.d_Variants, self.chrom)

    def __setstate__(self, state):
        (self.d_Variants, self.chrom) = state

    @cython.ccall
    def n_variants(self) -> cython.int:
        return len(self.d_Variants)

    @cython.ccall
    def add_variant(self, p: cython.long, v: Variant):
        self.d_Variants[p] = v

    @cython.ccall
    def has_indel(self) -> bool:
        p: cython.long

        for p in sorted(self.d_Variants.keys()):
            if self.d_Variants[p].is_indel():
                return True
        return False

    @cython.ccall
    def has_refer_biased_01(self) -> bool:
        p: cython.long

        for p in sorted(self.d_Variants.keys()):
            if self.d_Variants[p].is_refer_biased_01():
                return True
        return False

    @cython.ccall
    def get_refer_biased_01s(self) -> list:
        ret_poss: list = []
        p: cython.long

        for p in sorted(self.d_Variants.keys()):
            if self.d_Variants[p].is_refer_biased_01():
                ret_poss.append(p)
        return ret_poss

    @cython.ccall
    def remove_variant(self, p: cython.long):
        assert p in self.d_Variants
        self.d_Variants.pop(p)

    @cython.ccall
    def replace_variant(self, p: cython.long, v: Variant):
        assert p in self.d_Variants
        self.d_Variants[p] = v

    @cython.ccall
    def fix_indels(self):
        p0: cython.long
        p1: cython.long
        p: cython.long

        # merge continuous deletion
        p0 = -1                           #start of deletion chunk
        p1 = -1                           #end of deletion chunk
        for p in sorted(self.d_Variants.keys()):
            if p == p1+1 and self.d_Variants[p].is_only_del() and self.d_Variants[p0].is_only_del():
                # we keep p0, remove p, and add p's ref_allele to p0, keep other information as in p0
                if self.d_Variants[p0].top1isreference:
                    if self.d_Variants[p0]["top1allele"] == "*":
                        self.d_Variants[p0]["top1allele"] = ""
                    self.d_Variants[p0]["top1allele"] += self.d_Variants[p]["ref_allele"]
                elif self.d_Variants[p0].top2isreference:
                    if self.d_Variants[p0]["top2allele"] == "*":
                        self.d_Variants[p0]["top2allele"] = ""
                    self.d_Variants[p0]["top2allele"] += self.d_Variants[p]["ref_allele"]
                self.d_Variants[p0]["ref_allele"] += self.d_Variants[p]["ref_allele"]
                self.d_Variants.pop(p)
                p1 = p
            else:
                p0 = p
                p1 = p

        # fix deletion so that if the preceding base is 0/0 -- i.e. not in d_Variants, the reference base will be added.
        for p in sorted(self.d_Variants.keys()):
            if self.d_Variants[p].is_only_del():
                if not ((p-1) in self.d_Variants):
                    if p > self.start:  # now add the reference base
                        self.d_Variants[p-1] = copy(self.d_Variants[p])
                        rs = str(self.refseq)
                        self.d_Variants[p-1]["ref_allele"] = rs[p - self.start] + self.d_Variants[p-1]["ref_allele"]
                        self.d_Variants[p-1]["alt_allele"] = rs[p - self.start]
                        if self.d_Variants[p].top1isreference:
                            self.d_Variants[p-1]["top1allele"] = self.d_Variants[p-1]["ref_allele"]
                            self.d_Variants[p-1]["top2allele"] = self.d_Variants[p-1]["alt_allele"]
                        elif self.d_Variants[p].top2isreference:
                            self.d_Variants[p-1]["top1allele"] = self.d_Variants[p-1]["alt_allele"]
                            self.d_Variants[p-1]["top2allele"] = self.d_Variants[p-1]["ref_allele"]
                        self.d_Variants.pop(p)

        # remove indel if a deletion is immediately following an
        # insertion -- either a third genotype is found which is not
        # allowed in this version of sapper, or problem caused by
        # assembling in a simple repeat region.
        for p in sorted(self.d_Variants.keys()):
            if self.d_Variants[p].is_only_del():
                if (p-1) in self.d_Variants and self.d_Variants[p-1].is_only_insertion():
                    self.d_Variants.pop(p)
                    self.d_Variants.pop(p - 1)
        return

    @cython.ccall
    def toVCF(self) -> str:
        p: cython.long
        res: str

        res = ""
        for p in sorted(self.d_Variants.keys()):
            res += "\t".join((self.chrom, str(p+1), ".", self.d_Variants[p].toVCF())) + "\n"
        return res



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2025-07-24 15:46:59 Tao Liu>

"""Module description:

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""
# ------------------------------------
# python modules
# ------------------------------------
from math import sqrt
import cython
import numpy as np
from cython.cimports import numpy as cnp
from MACS3.Signal.Prob import pnorm2
from MACS3.Signal.PairedEndTrack import PETrackI
from MACS3.Utilities.Logger import logging

logger = logging.getLogger(__name__)
debug = logger.debug
info = logger.info

# ------------------------------------
# Misc functions
# ------------------------------------


@cython.cfunc
def online_update(x: cython.float, c: cython.long,
                  m: cython.float, s: cython.float) -> cython.tuple:
    delta: cython.float

    c += 1
    if c == 1:
        return (1, x, float(0.0))
    delta = x - m
    m += delta / c
    s += delta * (x - m)
    return (c, m, s)


# https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.norm.html


@cython.inline
@cython.cfunc
def get_weighted_density(x: cython.int, m: cython.float, v:
                         cython.float, w: cython.float) -> cython.float:
    """Description:

    parameters:
      1. x: the observed value
      2. m: the mean of gaussian
      3. v: the variance of the gaussian
      4. w: the weight
    return value:
    """
    return w * pnorm2(float(x), m, v)


@cython.cfunc
def return_greater(data: cnp.ndarray) -> cython.int:
    """
    Return the index of the largest value in an array of doubles
    @param data: an Array of doubles
    @return an integer representing thre index of the largest value in the inputted array
    """

    largest_index: cython.int
    largest_inds: cnp.ndarray
    largest_inds = np.argwhere(data == np.amax(data))  # find the indices of the most likely category(s) (max values)
    if len(largest_inds) > 1:    # if there are multiple "most-likely" categories, ignore data point
        largest_index = -1
    else:
        largest_index = largest_inds[0][0]
    return largest_index

# ------------------------------------
# Classes
# ------------------------------------


@cython.cclass
class HMMR_EM:
    """ Main HMMR EM class.

    This EM trainer will find the optimal mean and stddev of three of
    the four modes -- mono-nucleosomal, di-nucloeosomal, and
    three-nucleosomal fragments. Please note that the mean and stddev
    of the short fragment won't be optimized in this approach and only
    rely on the user's input.

    """
    # fragment length mean for each of the three modes
    fragMeans = cython.declare(cnp.ndarray, visibility='public')
    # fragment length standard deviation for each of the three modes
    fragStddevs = cython.declare(cnp.ndarray, visibility='public')
    fragVars: cnp.ndarray  # we need variances to calculate PDF of normal distribution
    min_fraglen: cython.int
    max_fraglen: cython.int
    epsilon: cython.float  # maximum difference to call the value is converged
    maxIter: cython.int  # maximum iteration
    jump: cython.float
    seed: cython.int            # random seed for downsampling
    converged: cython.bint      # wheter the EM is converged
    sample_percentage: cython.float
    __petrack: PETrackI         # PETrackI object
    __data: cnp.ndarray          # data for fragment lengths
    __weights: cnp.ndarray

    def __init__(self, petrack: PETrackI, init_means: cython.list,
                 init_stddevs: cython.list,
                 min_fraglen: cython.int = 100,
                 max_fraglen: cython.int = 1000,
                 sample_percentage: cython.float = 10,
                 epsilon: cython.float = 0.05,
                 maxIter: cython.int = 20,
                 jump: cython.float = 1.5, seed: cython.int = 12345):
        """Initialize HMMR_EM object. The first three parameters are required.

        parameters:
            1. petrack: a MACS3.Signal.PairedEndTrack.PETrackI object
            2. init_means: list of initial means of fragments, for mono, di, and tri signals
            3. init_stddevs: list of initial stddevs of fragments, for mono, di, and tri signals
            4. min_fraglen
            5. max_fraglen
            6. sample_percentage: downsample the original data to get the lengths distribution, default 10
            7. epsilon
            8. maxIter
            9. jump: to emplify the difference
            9. seed
        """
        cutoff1: cython.float
        cutoff2: cython.float
        sum1: cython.long
        sum2: cython.long
        sum3: cython.long
        counter1: cython.long
        counter2: cython.long
        counter3: cython.long

        # initial values
        self.min_fraglen = min_fraglen
        self.max_fraglen = max_fraglen
        self.epsilon = epsilon
        self.maxIter = maxIter
        self.jump = jump
        self.seed = seed
        self.converged = False
        self.fragMeans = np.array(init_means, dtype='f4')
        self.fragStddevs = np.array(init_stddevs, dtype='f4')
        self.fragVars = (np.array(init_stddevs, dtype='f4')) ** 2
        self.sample_percentage = sample_percentage

        # first, let's prepare the lengths data
        # sample down
        self.__petrack = petrack.sample_percent_copy(self.sample_percentage/100.0, seed=self.seed)  # may need to provide seed option for init function
        self.__data = self.__petrack.fraglengths()
        # fhd = open("allfrag.txt","w")
        # for d in self.__data:
        #    fhd.write(f"{d}\n")
        # fhd.close()
        # then we only keep those with fragment lengths within certain range
        self.__data = self.__data[np.logical_and(self.__data >= self.min_fraglen, self.__data <= self.max_fraglen)]

        info(f"# Downsampled {self.__data.shape[0]} fragments will be used for EM training...")
        # next, we will calculate the weights -- ie the proportion of fragments in each length category
        cutoff1 = (init_means[1] - init_means[0])/2 + init_means[0]
        cutoff2 = (init_means[2] - init_means[1])/2 + init_means[1]

        sum3 = len(self.__data)
        sum2 = sum(self.__data < cutoff2)
        sum1 = sum(self.__data < cutoff1)

        counter3 = sum3 - sum2
        counter2 = sum2 - sum1
        counter1 = sum1

        self.__weights = np.array([counter1/sum3, counter2/sum3, counter3/sum3])
        debug(f"initial: means: {self.fragMeans}, stddevs: {self.fragStddevs}, weights: {self.__weights}")
        self.__learn()
        return

    @cython.cfunc
    def __learn(self) -> cython.bint:
        """Description: When we train the mean and stddev

        parameters:
        return value:
        """
        itr: cython.int = 0
        i: cython.int
        counter: cython.int
        old_means: cnp.ndarray = np.zeros(3, dtype='f4')
        old_stddevs: cnp.ndarray = np.zeros(3, dtype='f4')
        old_weights: cnp.ndarray = np.zeros(3, dtype='f4')

        self.converged = False
        while self.converged is False:
            for i in range(3):       
                old_means[i] = self.fragMeans[i]
                old_stddevs[i] = self.fragStddevs[i]
                old_weights[i] = self.__weights[i]

            self.__iterate()
            itr += 1
            debug(f"after iteration {itr}: means: {self.fragMeans}, stddevs: {self.fragStddevs}, weights: {self.__weights}")
            counter = 0
            for i in range(3):
                if abs(old_means[i] - self.fragMeans[i]) < self.epsilon and abs(old_weights[i] - self.__weights[i]) < self.epsilon and abs(old_stddevs[i] - self.fragStddevs[i]) < self.epsilon:
                    counter += 1
            if counter == 3:
                self.converged = True
                info(f"# Reached convergence after {itr} iterations")
            if itr >= self.maxIter:
                info(f"# Reached maximum number ({self.maxIter}) of iterations")
                break
        return self.converged

    @cython.cfunc
    def __iterate(self):
        """Description: This is a private function only used by HMMR_EM class

        parameters:
        return value:
        """
        temp: cnp.ndarray
        means: cnp.ndarray
        variances: cnp.ndarray
        s: cnp.ndarray
        c: cnp.ndarray
        total: cython.long
        i: cython.int
        j: cython.int
        index: cython.int

        temp = np.zeros(3, dtype='f4')  # for each category, the likelihood
        total = 0  # total number of data points/fragments assigned to three categories
        means = np.zeros(3, dtype='f4')  # for each category, the new mean
        variances = np.zeros(3, dtype='f4')  # for each category, the new variances
        s = np.zeros(3, dtype='f4')          # for each category, __s
                                            # is for storing
                                            # intermediate values
                                            # for the online
                                            # algorithm
        c = np.zeros(3, dtype='i8')          # for each category, __c
                                            # is for storing
                                            # intermediate values for
                                            # the online algorithm
        for i in range(len(self.__data)):
            for j in range(3):
                # for each category: mono, di, tri- (3 in total), we get the likelihoods
                # debug( f"j:{j} {self.__data[i]}, {self.fragMeans[j]}, {self.fragVars[j]}, {self.__weights[j]}" )
                temp[j] = get_weighted_density(self.__data[i], self.fragMeans[j], self.fragVars[j], self.__weights[j])
            # now look for the most likely category, as `index`
            index = return_greater(temp)

            # then we will update __means and __stds
            if index != -1:  # If we can find a mostly likely category
                # #---- update with online algorithm --
                (c[index], means[index], s[index]) = online_update(self.__data[i], c[index], means[index], s[index])
                variances[index] = s[index]/c[index]
                total += 1
            # debug( f"index: {index} mean: {means[index]}" )

        # Set fragMeans, fragStddevs, and fragVars
        # debug( f"counts: {c[0]} {c[1]} {c[2]}" )
        # debug( f"means: {means[0]} {means[1]} {means[2]}" )
        # debug( f"vars: {variances[0]} {variances[1]} {variances[2]}")
        try:
            for j in range(3):
                if c[j] == 0:
                    # if there is no fragment size in this category, ignore
                    continue
                self.fragMeans[j] = self.fragMeans[j] + self.jump*(means[j] - self.fragMeans[j])
                self.fragVars[j] = self.fragVars[j] + self.jump*(variances[j] - self.fragVars[j])
                self.fragStddevs[j] = sqrt(self.fragVars[j])
                self.__weights[j] = c[j] / total
        except ValueError:
            print(' ValueError:  Adjust --means and --stddevs options and re-run command')
        debug(f"After this iteration, {total} fragments have been assigned with either of the three modes")
        return



cdef extern from "fml.h":
    ctypedef struct bseq1_t:
        int l_seq
        char *seq
        char *qual # NULL-terminated strings; length expected to match $l_seq

    ctypedef struct magopt_t:
        int flag, min_ovlp, min_elen, min_ensr, min_insr, max_bdist, max_bdiff, max_bvtx, min_merge_len, trim_len, trim_depth
        float min_dratio1, max_bcov, max_bfrac

    ctypedef struct fml_opt_t:
        int n_threads        # number of threads; don't use multi-threading for small data sets
        int ec_k             # k-mer length for error correction; 0 for auto estimate
        int min_cnt, max_cnt # both occ threshold in ec and tip threshold in cleaning lie in [min_cnt,max_cnt]
        int min_asm_ovlp     # min overlap length during assembly
        int min_merge_len    # during assembly, don't explicitly merge an overlap if shorter than this value
        magopt_t mag_opt     # graph cleaning options

    ctypedef struct fml_ovlp_t:
        unsigned int len_, from_, id_, to_ 
        #unit32_t from  # $from and $to: 0 meaning overlapping 5'-end; 1 overlapping 3'-end
        #unsigned int id
        #unsigned int to    # $id: unitig number

    ctypedef struct fml_utg_t:
        int len      # length of sequence
        int nsr      # number of supporting reads
        char *seq        # unitig sequence
        char *cov        # cov[i]-33 gives per-base coverage at i
        int n_ovlp[2]    # number of 5'-end [0] and 3'-end [1] overlaps
        fml_ovlp_t *ovlp # overlaps, of size n_ovlp[0]+n_ovlp[1]

    void fml_opt_init(fml_opt_t *opt)
    fml_utg_t* fml_assemble(const fml_opt_t *opt, int n_seqs, bseq1_t *seqs, int *n_utg)
    void fml_utg_destroy(int n_utg, fml_utg_t *utg)
    void fml_utg_print(int n_utgs, const fml_utg_t *utg)
    bseq1_t *bseq_read(const char *fn, int *n)

# --- end of fermi-lite functions ---

# --- smith-waterman alignment functions ---

cdef extern from "swalign.h":
    ctypedef struct seq_pair_t:
        char *a
        unsigned int alen
        char *b
        unsigned int blen
    ctypedef struct align_t:
        seq_pair_t *seqs
        char *markup;
        int start_a
        int start_b
        int end_a
        int end_b
        int matches
        int gaps
        double score
    align_t *smith_waterman(seq_pair_t *problem)
    void destroy_seq_pair(seq_pair_t *pair)
    void destroy_align(align_t *ali)



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2024-10-22 17:14:11 Tao Liu>

"""Module

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).
"""
# ------------------------------------
# python modules
# ------------------------------------
from operator import itemgetter

from MACS3.Signal.ReadAlignment import ReadAlignment
from MACS3.Signal.PosReadsInfo import PosReadsInfo
from MACS3.IO.PeakIO import PeakIO

import cython
from cython.cimports.cpython import bool

# ------------------------------------
# constants
# ------------------------------------
__version__ = "Parser $Revision$"
__author__ = "Tao Liu <tliu4@buffalo.edu>"
__doc__ = "All Parser classes"

__DNACOMPLEMENT__ = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f !"#$%&\'()*+,-./0123456789:;<=>?@TBGDEFCHIJKLMNOPQRSAUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff' # A trans table to convert A to T, C to G, G to C, and T to A.

__CIGARCODE__ = "MIDNSHP=X"

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Classes
# ------------------------------------


@cython.cclass
class UnitigRAs:
    """
    """
    RAlists: list            # [RAlists_T, RAlists_C]
    seq: bytes
    unitig_aln: bytes
    reference_aln: bytes
    chrom: bytes
    lpos: cython.long
    rpos: cython.long
    unitig_length: cython.long
    reference_length: cython.long
    aln_length: cython.long

    def __init__(self,
                 chrom: bytes,
                 lpos: cython.long,
                 rpos: cython.long,
                 unitig_aln: bytes,
                 reference_aln: bytes,
                 RAlists: list):
        assert len(unitig_aln) == len(reference_aln), Exception("aln on unitig and reference should be the same length!")
        self.chrom = chrom
        self.lpos = lpos
        self.rpos = rpos
        self.unitig_aln = unitig_aln
        self.reference_aln = reference_aln
        self.RAlists = RAlists
        # fill in other information
        self.seq = self.unitig_aln.replace(b'-', b'')
        self.unitig_length = len(self.seq)
        self.reference_length = rpos - lpos
        self.aln_length = len(unitig_aln)

    def __getitem__(self, keyname):
        if keyname == "chrom":
            return self.chrom
        elif keyname == "lpos":
            return self.lpos
        elif keyname == "rpos":
            return self.rpos
        elif keyname == "seq":
            return self.seq
        elif keyname == "unitig_aln":
            return self.unitig_aln
        elif keyname == "reference_aln":
            return self.reference_aln
        elif keyname == "unitig_length":
            return self.unitig_length
        elif keyname == "reference_length":
            return self.reference_length
        elif keyname == "aln_length":
            return self.aln_length
        elif keyname == "count":
            return len(self.RAlists[0]) + len(self.RAlists[1])
        else:
            raise KeyError("Unavailable key:", keyname)

    def __getstate__(self):
        return (self.RAlists, self.seq, self.unitig_aln, self.reference_aln,
                self.chrom, self.lpos, self.rpos, self.unitig_length,
                self.reference_length, self.aln_length)

    def __setstate__(self, state):
        (self.RAlists, self.seq, self.unitig_aln, self.reference_aln,
         self.chrom, self.lpos, self.rpos, self.unitig_length,
         self.reference_length, self.aln_length) = state

    @cython.ccall
    def get_variant_bq_by_ref_pos(self,
                                  ref_pos: cython.long) -> tuple:
        """
        return (s, bq_list_t, bq_list_c, strand_list_t, strand_list_c)
        """
        i: cython.long
        index_aln: cython.long
        index_unitig: cython.long
        residue: cython.long
        ra: ReadAlignment
        s: bytes
        bq_list_t: list = []
        bq_list_c: list = []
        strand_list_t: list = []
        strand_list_c: list = []
        tip_list_t: list = []
        pos_list_t: list = []
        pos_list_c: list = []
        ra_seq: bytes
        ra_pos: cython.long
        l_read: cython.int

        #  b'TTATTAGAAAAAAT' find = 2
        #          b'AAAAATCCCACAGG'
        # b'TTTTATTAGAAAAAATCCCACAGGCAGCCACTAGGTGGCAGTAACAGGCTTTTGCCAGCGGCTCCAGTCAGCATGGCTTGACTGTGTGCT'
        # b'TTTTATTACAAAAA-TCCCACAGGCAGCCACTAGGTGGCAGTAACAGGCTTTTGCCAGCGGCTCCAGTCAGCATGGCTTGACTGTGTGCT' lpos=100
        #          |    |       |
        # genome   108  113     120
        # aln       8    13      21
        # unitig    8    13      21
        # ref       8    13      20
        # read1     6    11
        # read2           3      11
        # find the position
        residue = ref_pos - self.lpos + 1
        index_aln = 0
        for i in range(self.aln_length):
            if self.reference_aln[i] != 45:  # 45 means b'-'
                residue -= 1
            if residue == 0:
                break
            index_aln += 1

        # index_aln should be the position on aln
        s = self.unitig_aln[index_aln:index_aln+1]
        # find the index on unitig
        index_unitig = len(self.unitig_aln[:index_aln+1].replace(b'-', b''))

        if s == b'-':                     # deletion
            for ra in self.RAlists[0]:
                ra_seq = ra["SEQ"]
                l_read = ra["l"]
                ra_pos = index_unitig - self.seq.find(ra_seq) - 1
                if ra_pos == 0 or ra_pos == l_read - 1:
                    tip_list_t.append(True)
                else:
                    tip_list_t.append(False)
                bq_list_t.append(93)
                strand_list_t.append(ra["strand"])
                pos_list_t.append(ra_pos)
            for ra in self.RAlists[1]:
                ra_seq = ra["SEQ"]
                ra_pos = index_unitig - self.seq.find(ra_seq) - 1
                bq_list_c.append(93)
                strand_list_c.append(ra["strand"])
                pos_list_c.append(ra_pos)
            return (bytearray(b'*'), bq_list_t, bq_list_c, strand_list_t,
                    strand_list_c, tip_list_t, pos_list_t, pos_list_c)

        if index_aln < self.aln_length - 1:
            for i in range(index_aln + 1, self.aln_length):
                if self.reference_aln[i] == 45:   # insertion detected, 45 means b'-'
                    s += self.unitig_aln[i:i+1]  # we extend the s string to contain the inserted seq
                else:
                    break

        for ra in self.RAlists[0]:        # treatment
            ra_seq = ra["SEQ"]
            l_read = ra["l"]
            ra_pos = index_unitig - self.seq.find(ra_seq) - 1
            if ra_pos < l_read and ra_pos >= 0:
                pos_list_t.append(ra_pos)
                if ra_pos == 0 or ra_pos == l_read - 1:
                    tip_list_t.append(True)
                else:
                    tip_list_t.append(False)
                bq_list_t.append(ra["binaryqual"][ra_pos])
                strand_list_t.append(ra["strand"])

        for ra in self.RAlists[1]:        # control
            ra_seq = ra["SEQ"]
            l_read = ra["l"]
            ra_pos = index_unitig - self.seq.find(ra_seq) - 1
            if ra_pos < l_read and ra_pos >= 0:
                pos_list_c.append(ra_pos)
                bq_list_c.append(ra["binaryqual"][ra_pos])
                strand_list_c.append(ra["strand"])

        return (bytearray(s), bq_list_t, bq_list_c, strand_list_t,
                strand_list_c, tip_list_t, pos_list_t, pos_list_c)


@cython.cclass
class UnitigCollection:
    """A collection of ReadAlignment objects and the corresponding
    PeakIO.

    """
    chrom: bytes
    peak: PeakIO             # A PeakIO object
    URAs_list: list
    left: cython.long               # left position of peak
    right: cython.long              # right position of peak
    length: cython.long             # length of peak
    URAs_left: cython.long          # left position of all RAs in the collection
    URAs_right: cython.long         # right position of all RAs in the collection
    is_sorted: bool             # if sorted by lpos

    def __init__(self,
                 chrom: bytes,
                 peak: PeakIO,
                 URAs_list: list = []):
        self.chrom = chrom
        self.peak = peak
        self.URAs_list = URAs_list
        self.left = peak["start"]
        self.right = peak["end"]
        self.length = self.right - self.left
        self.URAs_left = URAs_list[0]["lpos"]  # initial assignment of RAs_left
        self.URAs_right = URAs_list[-1]["rpos"]  # initial assignment of RAs_right
        self.sort()                           # it will set self.is_sorted = True
        # check RAs_left and RAs_right
        for ura in URAs_list:
            if ura["lpos"] < self.URAs_left:
                self.URAs_left = ura["lpos"]
            if ura["rpos"] > self.URAs_right:
                self.URAs_right = ura["rpos"]

    def __getitem__(self, keyname):
        if keyname == "chrom":
            return self.chrom
        elif keyname == "left":
            return self.left
        elif keyname == "right":
            return self.right
        elif keyname == "URAs_left":
            return self.URAs_left
        elif keyname == "URAs_right":
            return self.URAs_right
        elif keyname == "length":
            return self.length
        elif keyname == "count":
            return len(self.URAs_list)
        elif keyname == "URAs_list":
            return self.URAs_list
        else:
            raise KeyError("Unavailable key:", keyname)

    def __getstate__(self):
        return (self.chrom, self.peak, self.URAs_list, self.left, self.right,
                self.length, self.URAs_left, self.URAs_right, self.is_sorted)

    def __setstate__(self, state):
        (self.chrom, self.peak, self.URAs_list, self.left, self.right,
         self.length, self.URAs_left, self.URAs_right, self.is_sorted) = state

    @cython.ccall
    def sort(self):
        """Sort RAs according to lpos. Should be used after realignment.

        """
        self.URAs_list.sort(key=itemgetter("lpos"))
        self.is_sorted = True
        return

    @cython.ccall
    def get_PosReadsInfo_ref_pos(self,
                                 ref_pos: cython.long,
                                 ref_nt: bytes,
                                 Q: cython.int = 20):
        """Generate a PosReadsInfo for: object a given reference genome
        position.

        Return a PosReadsInfo object.

        """
        s: bytearray
        bq_list_t: list
        bq_list_c: list
        strand_list_t: list
        strand_list_c: list
        tip_list_t: list
        pos_list_t: list
        pos_list_c: list
        ura: object
        i: cython.int
        posreadsinfo_p: PosReadsInfo

        posreadsinfo_p = PosReadsInfo(ref_pos, ref_nt)
        for i in range(len(self.URAs_list)):
            ura = self.URAs_list[i]
            if ura["lpos"] <= ref_pos and ura["rpos"] > ref_pos:
                (s, bq_list_t, bq_list_c, strand_list_t, strand_list_c,
                 tip_list_t, pos_list_t, pos_list_c) = ura.get_variant_bq_by_ref_pos(ref_pos)
                for i in range(len(bq_list_t)):
                    posreadsinfo_p.add_T(i, bytes(s), bq_list_t[i],
                                         strand_list_t[i], tip_list_t[i], Q=Q)
                for i in range(len(bq_list_c)):
                    posreadsinfo_p.add_C(i, bytes(s), bq_list_c[i],
                                         strand_list_c[i], Q=Q)

        return posreadsinfo_p



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2025-11-20 10:58:03 Tao Liu>

"""Scoring utilities for MACS3 signal tracks and peak callers.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
from functools import reduce

# ------------------------------------
# MACS3 modules
# ------------------------------------
from MACS3.Signal.SignalProcessing import maxima, enforce_peakyness
from MACS3.Signal.Prob import poisson_cdf
from MACS3.IO.PeakIO import PeakIO, BroadPeakIO

# ------------------------------------
# Other modules
# ------------------------------------
import cython
import numpy as np
import cython.cimports.numpy as cnp
from cython.cimports.cpython import bool
from cykhash import PyObjectMap, Float32to32Map

# ------------------------------------
# C lib
# ------------------------------------
from cython.cimports.libc.math import (log10,
                                       log)

# ------------------------------------
# constants
# ------------------------------------

# ------------------------------------
# Misc functions
# ------------------------------------


@cython.inline
@cython.cfunc
def int_max(a: cython.int, b: cython.int) -> cython.int:
    """Return the larger of ``a`` and ``b``."""
    return a if a >= b else b


@cython.inline
@cython.cfunc
def int_min(a: cython.int, b: cython.int) -> cython.int:
    """Return the smaller of ``a`` and ``b``."""
    return a if a <= b else b


LOG10_E: cython.float = 0.43429448190325176

pscore_dict = PyObjectMap()


@cython.cfunc
def get_pscore(observed: cython.int,
               expectation: cython.float) -> cython.float:
    """Return cached ``-log10`` Poisson tail probability for ``observed``."""
    score: cython.double

    try:
        return pscore_dict[(observed, expectation)]
    except KeyError:
        score = -1 * poisson_cdf(observed,
                                 expectation,
                                 False,
                                 True)
        pscore_dict[(observed, expectation)] = score
        return score


asym_logLR_dict = PyObjectMap()


@cython.cfunc
def logLR_asym(x: cython.float,
               y: cython.float) -> cython.float:
    """Return asymmetric ``log10`` likelihood ratio between ``x`` and ``y``."""
    s: cython.float

    if (x, y) in asym_logLR_dict:
        return asym_logLR_dict[(x, y)]
    else:
        if x > y:
            s = (x*(log(x)-log(y))+y-x)*LOG10_E
        elif x < y:
            s = (x*(-log(x)+log(y))-y+x)*LOG10_E
        else:
            s = 0
        asym_logLR_dict[(x, y)] = s
        return s


sym_logLR_dict = PyObjectMap()


@cython.cfunc
def logLR_sym(x: cython.float, y: cython.float) -> cython.float:
    """Return symmetric ``log10`` likelihood ratio between ``x`` and ``y``."""
    s: cython.float

    if (x, y) in sym_logLR_dict:
        return sym_logLR_dict[(x, y)]
    else:
        if x > y:
            s = (x*(log(x)-log(y))+y-x)*LOG10_E
        elif y > x:
            s = (y*(log(x)-log(y))+y-x)*LOG10_E
        else:
            s = 0
        sym_logLR_dict[(x, y)] = s
        return s


@cython.inline
@cython.cfunc
def get_logFE(x: cython.float, y: cython.float) -> cython.float:
    """Return ``log10`` fold enrichment (base-10) for ``x`` over ``y``."""
    return log10(x/y)


@cython.cfunc
def get_subtraction(x: cython.float, y: cython.float) -> cython.float:
    """Return the difference ``x - y``."""
    return x - y

# ------------------------------------
# Classes
# ------------------------------------


@cython.cclass
class ScoreTrackII:
    """Container for treatment/control pileups and derived score tracks."""
    # dictionary for data of each chromosome
    data: dict
    # length of data array of each chromosome
    datalength: dict
    # whether trackline should be saved in bedGraph
    trackline: bool
    # seq depth in million of treatment
    treat_edm: cython.float
    # seq depth in million of control
    ctrl_edm: cython.float
    # method for calculating scores.
    scoring_method: cython.char
    # scale to control? scale to treatment? both scale to 1million reads?
    normalization_method: cython.char
    # the pseudocount used to calcuate logLR, FE or logFE
    pseudocount: cython.float
    # cutoff
    cutoff: cython.float
    # save pvalue<->length dictionary
    pvalue_stat: dict

    def __init__(self,
                 treat_depth: cython.float,
                 ctrl_depth: cython.float,
                 pseudocount: cython.float = 1.0):
        """Initialise score containers with effective library depths.

        Args:
            treat_depth: Effective treatment depth (millions of filtered reads).
            ctrl_depth: Effective control depth (millions of filtered reads).
            pseudocount: Pseudocount added when computing score metrics.
        """
        # for each chromosome, there is a l*4 matrix. First column:
        # end position of a region; Second: treatment pileup; third:
        # control pileup ; forth: score (can be p/q-value/likelihood
        # ratio/fold-enrichment/subtraction depending on -c setting)
        self.data = {}

        self.datalength = {}
        self.trackline = False
        self.treat_edm = treat_depth
        self.ctrl_edm = ctrl_depth

        # scoring_method:  p: -log10 pvalue;
        #                  q: -log10 qvalue;
        #                  l: log10 likelihood ratio (minus for depletion)
        #                  f: log10 fold enrichment
        #                  F: linear fold enrichment
        #                  d: subtraction
        #                  m: fragment pileup per million reads
        #                  N: not set
        self.scoring_method = ord("N")

        # normalization_method: T: scale to depth of treatment;
        #                       C: scale to depth of control;
        #                       M: scale to depth of 1 million;
        #                       N: not set/ raw pileup
        self.normalization_method = ord("N")

        self.pseudocount = pseudocount
        self.pvalue_stat = {}

    @cython.ccall
    def set_pseudocount(self, pseudocount: cython.float):
        """Update the pseudocount used when computing score metrics."""
        self.pseudocount = pseudocount

    @cython.ccall
    def enable_trackline(self):
        """Enable UCSC track line output when exporting bedGraphs."""
        self.trackline = True

    @cython.ccall
    def add_chromosome(self,
                       chrom: bytes,
                       chrom_max_len: cython.int):
        """Allocate arrays for ``chrom`` with capacity ``chrom_max_len``."""
        if chrom not in self.data:
            self.data[chrom] = [np.zeros(chrom_max_len, dtype="int32"),  # pos
                                # pileup at each interval, in float32 format
                                np.zeros(chrom_max_len, dtype="float32"),
                                # control at each interval, in float32 format
                                np.zeros(chrom_max_len, dtype="float32"),
                                # score at each interval, in float32 format
                                np.zeros(chrom_max_len, dtype="float32")]
            self.datalength[chrom] = 0

    @cython.ccall
    def add(self,
            chromosome: bytes,
            endpos: cython.int,
            chip: cython.float,
            control: cython.float):
        """Append treatment/control pileup ending at ``endpos`` for ``chromosome``."""
        i: cython.int

        i = self.datalength[chromosome]
        c = self.data[chromosome]
        c[0][i] = endpos
        c[1][i] = chip
        c[2][i] = control
        self.datalength[chromosome] += 1

    @cython.ccall
    def finalize(self):
        """Trim per-chromosome arrays to their populated length."""
        chrom: bytes
        ln: cython.int

        for chrom in sorted(self.data.keys()):
            d = self.data[chrom]
            ln = self.datalength[chrom]
            d[0].resize(ln, refcheck=False)
            d[1].resize(ln, refcheck=False)
            d[2].resize(ln, refcheck=False)
            d[3].resize(ln, refcheck=False)
        return

    @cython.ccall
    def get_data_by_chr(self,
                        chromosome: bytes):
        """Return ``(positions, treatment, control, score)`` arrays for ``chromosome``."""
        if chromosome in self.data:
            return self.data[chromosome]
        else:
            return None

    @cython.ccall
    def get_chr_names(self):
        """Return all the chromosome names stored.

        """
        return set(self.data.keys())

    @cython.ccall
    def change_normalization_method(self,
                                    normalization_method: cython.char):
        """Change/set normalization method. However, I do not
        recommend change this back and forward, since some precision
        issue will happen -- I only keep two digits.

        normalization_method: T: scale to depth of treatment;
                              C: scale to depth of control;
                              M: scale to depth of 1 million;
                              N: not set/ raw pileup
        """
        if normalization_method == ord('T'):
            if self.normalization_method == ord('T'):  # do nothing
                pass
            elif self.normalization_method == ord('C'):
                self.normalize(self.treat_edm/self.ctrl_edm,
                               self.treat_edm/self.ctrl_edm)
            elif self.normalization_method == ord('M'):
                self.normalize(self.treat_edm, self.treat_edm)
            elif self.normalization_method == ord('N'):
                self.normalize(1, self.treat_edm/self.ctrl_edm)
            else:
                raise NotImplementedError
            self.normalization_method = ord('T')
        elif normalization_method == ord('C'):
            if self.normalization_method == ord('T'):
                self.normalize(self.ctrl_edm/self.treat_edm,
                               self.ctrl_edm/self.treat_edm)
            elif self.normalization_method == ord('C'):  # do nothing
                pass
            elif self.normalization_method == ord('M'):
                self.normalize(self.ctrl_edm, self.ctrl_edm)
            elif self.normalization_method == ord('N'):
                self.normalize(self.ctrl_edm/self.treat_edm, 1)
            else:
                raise NotImplementedError
            self.normalization_method = ord('C')
        elif normalization_method == ord('M'):
            if self.normalization_method == ord('T'):
                self.normalize(1/self.treat_edm,
                               1/self.treat_edm)
            elif self.normalization_method == ord('C'):
                self.normalize(1/self.ctrl_edm,
                               1/self.ctrl_edm)
            elif self.normalization_method == ord('M'):  # do nothing
                pass
            elif self.normalization_method == ord('N'):
                self.normalize(1/self.treat_edm,
                               1/self.ctrl_edm)
            else:
                raise NotImplementedError
            self.normalization_method = ord('M')
        elif normalization_method == ord('N'):
            if self.normalization_method == ord('T'):
                self.normalize(self.treat_edm,
                               self.treat_edm)
            elif self.normalization_method == ord('C'):
                self.normalize(self.ctrl_edm,
                               self.ctrl_edm)
            elif self.normalization_method == ord('M'):
                self.normalize(self.treat_edm,
                               self.ctrl_edm)
            elif self.normalization_method == ord('N'):  # do nothing
                pass
            else:
                raise NotImplementedError
            self.normalization_method = ord('N')

    @cython.cfunc
    def normalize(self,
                  treat_scale: cython.float,
                  control_scale: cython.float):
        """Scale treatment and control pileups in-place by the given factors."""
        p: cnp.ndarray
        c: cnp.ndarray
        ln: cython.long
        i: cython.long

        for chrom in sorted(self.data.keys()):
            p = self.data[chrom][1]
            c = self.data[chrom][2]
            ln = self.datalength[chrom]
            for i in range(ln):
                p[i] *= treat_scale
                c[i] *= control_scale
        return

    @cython.ccall
    def change_score_method(self,
                            scoring_method: cython.char):
        """
        scoring_method:  p: -log10 pvalue;
                         q: -log10 qvalue;
                         l: log10 likelihood ratio (minus for depletion)
                         s: symmetric log10 likelihood ratio (for comparing two
                            ChIPs)
                         f: log10 fold enrichment
                         F: linear fold enrichment
                         d: subtraction
                         M: maximum
                         m: fragment pileup per million reads
        """
        if scoring_method == ord('p'):
            self.compute_pvalue()
        elif scoring_method == ord('q'):
            # if not already calculated p, compute pvalue first
            if self.scoring_method != ord('p'):
                self.compute_pvalue()
            self.compute_qvalue()
        elif scoring_method == ord('l'):
            self.compute_likelihood()
        elif scoring_method == ord('s'):
            self.compute_sym_likelihood()
        elif scoring_method == ord('f'):
            self.compute_logFE()
        elif scoring_method == ord('F'):
            self.compute_foldenrichment()
        elif scoring_method == ord('d'):
            self.compute_subtraction()
        elif scoring_method == ord('m'):
            self.compute_SPMR()
        elif scoring_method == ord('M'):
            self.compute_max()
        else:
            raise NotImplementedError

    @cython.cfunc
    def compute_pvalue(self):
        """Compute -log_{10}(pvalue)
        """
        p: cnp.ndarray
        c: cnp.ndarray
        v: cnp.ndarray
        pos: cnp.ndarray
        ln: cython.long
        tmp_l: cython.long
        i: cython.long
        prev_pos: cython.long
        chrom: bytes

        for chrom in sorted(self.data.keys()):
            prev_pos = 0
            pos = self.data[chrom][0]
            p = self.data[chrom][1]
            c = self.data[chrom][2]
            v = self.data[chrom][3]
            ln = self.datalength[chrom]
            for i in range(ln):
                v[i] = get_pscore(cython.cast(cython.int,
                                              (p[i] + self.pseudocount)),
                                  c[i] + self.pseudocount)
                tmp_l = pos[i] - prev_pos
                try:
                    self.pvalue_stat[v[i]] += tmp_l
                except Exception:
                    self.pvalue_stat[v[i]] = tmp_l
                prev_pos = pos[i]

        self.scoring_method = ord('p')
        return

    @cython.cfunc
    def compute_qvalue(self):
        """Compute -log_{10}(qvalue)
        """
        pqtable: object
        i: cython.long
        ln: cython.long
        chrom: bytes
        v: cnp.ndarray

        # pvalue should be computed first!
        assert self.scoring_method == ord('p')
        # make pqtable
        pqtable = self.make_pq_table()

        # convert p to q
        for chrom in sorted(self.data.keys()):
            v = self.data[chrom][3]
            ln = self.datalength[chrom]
            for i in range(ln):
                v[i] = pqtable[v[i]]

        self.scoring_method = ord('q')
        return

    @cython.ccall
    def make_pq_table(self):
        """Make pvalue-qvalue table.

        Step1: get all pvalue and length of block with this pvalue
        Step2: Sort them
        Step3: Apply AFDR method to adjust pvalue and get qvalue for
               each pvalue

        Return a dictionary of
        {-log10pvalue:(-log10qvalue,rank,basepairs)} relationships.

        """
        ln: cython.long
        i: cython.long
        j: cython.long
        v: cython.float
        q: cython.float
        pre_q: cython.float     # store the p and q scores
        N: cython.long
        k: cython.float
        f: cython.float
        pvalue2qvalue: object
        pvalue_stat: dict
        unique_values: list

        assert self.scoring_method == ord('p')

        pvalue_stat = self.pvalue_stat

        N = sum(pvalue_stat.values())
        k = 1                           # rank
        f = -log10(N)
        pre_q = 2147483647              # save the previous q-value

        pvalue2qvalue = Float32to32Map(for_int=False)
        unique_values = sorted(list(pvalue_stat.keys()), reverse=True)
        for i in range(len(unique_values)):
            v = unique_values[i]
            ln = pvalue_stat[v]
            q = v + (log10(k) + f)
            if q > pre_q:
                q = pre_q
            if q <= 0:
                q = 0
                break
            pvalue2qvalue[v] = q
            pre_q = q
            k += ln
        # bottom rank pscores all have qscores 0
        for j in range(i, len(unique_values)):
            v = unique_values[j]
            pvalue2qvalue[v] = 0
        return pvalue2qvalue

    @cython.cfunc
    def compute_likelihood(self):
        """Calculate log10 likelihood.

        """
        ln: cython.long
        i: cython.long
        chrom: bytes
        v1: cython.float
        v2: cython.float
        pseudocount: cython.float

        pseudocount = self.pseudocount

        for chrom in sorted(self.data.keys()):
            p = self.data[chrom][1].flat.__next__  # pileup in treatment
            c = self.data[chrom][2].flat.__next__  # pileup in control
            v = self.data[chrom][3]                # score
            ln = self.datalength[chrom]
            v1 = 2
            v2 = 1
            for i in range(ln):
                v1 = p()
                v2 = c()
                v[i] = logLR_asym(v1 + pseudocount, v2 + pseudocount)
        self.scoring_method = ord('l')
        return

    @cython.cfunc
    def compute_sym_likelihood(self):
        """Calculate symmetric log10 likelihood.

        """
        ln: cython.long
        i: cython.long
        chrom: bytes
        v1: cython.float
        v2: cython.float
        pseudocount: cython.float

        pseudocount = self.pseudocount

        for chrom in sorted(self.data.keys()):
            p = self.data[chrom][1].flat.__next__
            c = self.data[chrom][2].flat.__next__
            v = self.data[chrom][3]
            ln = self.datalength[chrom]
            v1 = 2
            v2 = 1
            for i in range(ln):
                v1 = p()
                v2 = c()
                v[i] = logLR_sym(v1 + pseudocount, v2 + pseudocount)
        self.scoring_method = ord('s')
        return

    @cython.cfunc
    def compute_logFE(self):
        """Calculate log10 fold enrichment (with 1 pseudocount).

        """
        p: cnp.ndarray
        c: cnp.ndarray
        v: cnp.ndarray
        ln: cython.long
        i: cython.long
        pseudocount: cython.float

        pseudocount = self.pseudocount

        for chrom in sorted(self.data.keys()):
            p = self.data[chrom][1]
            c = self.data[chrom][2]
            v = self.data[chrom][3]
            ln = self.datalength[chrom]
            for i in range(ln):
                v[i] = get_logFE(p[i] + pseudocount, c[i] + pseudocount)
        self.scoring_method = ord('f')
        return

    @cython.cfunc
    def compute_foldenrichment(self):
        """Calculate linear scale fold enrichment (with 1 pseudocount).

        """
        p: cnp.ndarray
        c: cnp.ndarray
        v: cnp.ndarray
        ln: cython.long
        i: cython.long
        pseudocount: cython.float

        pseudocount = self.pseudocount

        for chrom in sorted(self.data.keys()):
            p = self.data[chrom][1]
            c = self.data[chrom][2]
            v = self.data[chrom][3]
            ln = self.datalength[chrom]
            for i in range(ln):
                v[i] = (p[i] + pseudocount)/(c[i] + pseudocount)
        self.scoring_method = ord('F')
        return

    @cython.cfunc
    def compute_subtraction(self):
        """Populate scores with treatment minus control pileup."""
        p: cnp.ndarray
        c: cnp.ndarray
        v: cnp.ndarray
        ln: cython.long
        i: cython.long

        for chrom in sorted(self.data.keys()):
            p = self.data[chrom][1]
            c = self.data[chrom][2]
            v = self.data[chrom][3]
            ln = self.datalength[chrom]
            for i in range(ln):
                v[i] = p[i] - c[i]
        self.scoring_method = ord('d')
        return

    @cython.cfunc
    def compute_SPMR(self):
        """Populate scores with treatment pileup per million reads."""
        p: cnp.ndarray
        v: cnp.ndarray
        ln: cython.long
        i: cython.long
        scale: cython.float

        if self.normalization_method == ord('T') or self.normalization_method == ord('N'):
            scale = self.treat_edm
        elif self.normalization_method == ord('C'):
            scale = self.ctrl_edm
        elif self.normalization_method == ord('M'):
            scale = 1

        for chrom in sorted(self.data.keys()):
            p = self.data[chrom][1]
            v = self.data[chrom][3]
            ln = self.datalength[chrom]
            for i in range(ln):
                v[i] = p[i] / scale  # two digit precision may not be enough...
        self.scoring_method = ord('m')
        return

    @cython.cfunc
    def compute_max(self):
        """Populate scores with the element-wise maximum of treatment and control."""
        p: cnp.ndarray
        c: cnp.ndarray
        v: cnp.ndarray
        ln: cython.long
        i: cython.long

        for chrom in sorted(self.data.keys()):
            p = self.data[chrom][1]
            c = self.data[chrom][2]
            v = self.data[chrom][3]
            ln = self.datalength[chrom]
            for i in range(ln):
                v[i] = max(p[i], c[i])
        self.scoring_method = ord('M')
        return

    @cython.ccall
    def write_bedGraph(self,
                       fhd,
                       name: str,
                       description: str,
                       column: cython.short = 3):
        """Write all data to fhd in bedGraph Format.

        fhd: a filehandler to save bedGraph.

        name/description: the name and description in track line.

        colname: can be 1: chip, 2: control, 3: score

        """
        chrom: bytes
        ln: cython.int
        pre: cython.int
        i: cython.int
        p: cython.int
        pre_v: cython.float
        v: cython.float
        chrs: set
        pos: cnp.ndarray
        value: cnp.ndarray

        assert column in range(1, 4), "column should be between 1, 2 or 3."

        write = fhd.write

        if self.trackline:
            # this line is REQUIRED by the wiggle format for UCSC browser
            write("track type=bedGraph name=\"%s\" description=\"%s\"\n" %
                  (name.decode(), description))

        chrs = self.get_chr_names()
        for chrom in sorted(chrs):
            pos = self.data[chrom][0]
            value = self.data[chrom][column]
            ln = self.datalength[chrom]
            pre = 0
            if pos.shape[0] == 0:
                continue  # skip if there's no data
            pre_v = value[0]
            for i in range(1, ln):
                v = value[i]
                p = pos[i-1]
                if abs(pre_v - v) > 1e-5:  # precision is 5 digits
                    write("%s\t%d\t%d\t%.5f\n" %
                          (chrom.decode(), pre, p, pre_v))
                    pre_v = v
                    pre = p
            p = pos[-1]
            # last one
            write("%s\t%d\t%d\t%.5f\n" %
                  (chrom.decode(), pre, p, pre_v))

        return True

    @cython.ccall
    def cutoff_analysis(self,
                        max_gap: cython.int = 50,
                        min_length: cython.int = 200,
                        steps: cython.int = 100,
                        min_score: cython.float = 0,
                        max_score: cython.float = 1000) -> str:
        """Summarise peak metrics across a range of score thresholds.

        Args:
            max_gap: Maximum distance between merged regions.
            min_length: Minimum peak length to keep.
            steps: Number of cutoff increments between the observed
                minimum and maximum scores.
            min_score: Lower bound for the cutoff sweep.
            max_score: Upper bound for the cutoff sweep.

        Returns:
            str: Tab-delimited report of peak counts and lengths per cutoff.
        """
        chrs: set
        peak_content: list
        ret_list: list
        cutoff_list: list
        cutoff_npeaks: list
        cutoff_lpeaks: list
        chrom: bytes
        ret: str
        cutoff: cython.float
        total_l: cython.long
        total_p: cython.long
        i: cython.long
        n: cython.long
        ts: cython.long
        te: cython.long
        lastp: cython.long
        tl: cython.long
        peak_length: cython.long
        s: cython.float
        ln: cython.long
        chrom_min: cython.float
        chrom_max: cython.float
        obs_min: cython.float
        obs_max: cython.float
        minv: cython.float
        maxv: cython.float

        chrs = self.get_chr_names()

        obs_min = 10000000.0
        obs_max = -10000000.0
        for chrom in chrs:
            ln = self.datalength[chrom]
            if ln == 0:
                continue
            chrom_scores = self.data[chrom][3][:ln]
            if chrom_scores.size == 0:
                continue
            chrom_min = float(np.min(chrom_scores))
            chrom_max = float(np.max(chrom_scores))
            if chrom_min < obs_min:
                obs_min = chrom_min
            if chrom_max > obs_max:
                obs_max = chrom_max

        minv = max(min_score, obs_min)
        maxv = min(obs_max, max_score)
        if steps <= 0 or maxv <= minv:
            return "score\tnpeaks\tlpeaks\tavelpeak\n"

        s = float(maxv - minv)/steps
        if s <= 0:
            return "score\tnpeaks\tlpeaks\tavelpeak\n"

        cutoff_list = [round(value, 3) for value in np.arange(minv, maxv, s)]
        if not cutoff_list:
            return "score\tnpeaks\tlpeaks\tavelpeak\n"

        cutoff_npeaks = [0] * len(cutoff_list)
        cutoff_lpeaks = [0] * len(cutoff_list)

        for chrom in sorted(chrs):
            ln = self.datalength[chrom]
            if ln == 0:
                continue
            pos_array = self.data[chrom][0][:ln]
            score_array = self.data[chrom][3][:ln]

            for n in range(len(cutoff_list)):
                cutoff = cutoff_list[n]
                total_l = 0           # total length of peaks
                total_p = 0           # total number of peaks

                # get the regions with scores above cutoffs. This is
                # not an optimized method. It would be better to store
                # score array in a 2-D ndarray?
                above_cutoff = np.nonzero(score_array > cutoff)[0]
                # end positions of regions where score is above cutoff
                above_cutoff_endpos = pos_array[above_cutoff]
                # start positions of regions where score is above cutoff
                above_cutoff_startpos = pos_array[above_cutoff-1]

                if above_cutoff_endpos.size == 0:
                    continue

                # first bit of region above cutoff
                acs_next = iter(above_cutoff_startpos).__next__
                ace_next = iter(above_cutoff_endpos).__next__

                ts = acs_next()
                te = ace_next()
                peak_content = [(ts, te),]
                lastp = te

                for i in range(1, above_cutoff_startpos.size):
                    ts = acs_next()
                    te = ace_next()
                    tl = ts - lastp
                    if tl <= max_gap:
                        peak_content.append((ts, te))
                    else:
                        peak_length = peak_content[-1][1] - peak_content[0][0]
                        # if the peak is too small, reject it
                        if peak_length >= min_length:
                            total_l += peak_length
                            total_p += 1
                        peak_content = [(ts, te),]
                    lastp = te

                if peak_content:
                    peak_length = peak_content[-1][1] - peak_content[0][0]
                    # if the peak is too small, reject it
                    if peak_length >= min_length:
                        total_l += peak_length
                        total_p += 1
                cutoff_lpeaks[n] += total_l
                cutoff_npeaks[n] += total_p

        # prepare the returnning text
        ret_list = ["score\tnpeaks\tlpeaks\tavelpeak\n"]
        for n in range(len(cutoff_list)-1, -1, -1):
            cutoff = cutoff_list[n]
            if cutoff_npeaks[n] > 0:
                ret_list.append("%.2f\t%d\t%d\t%.2f\n" % (cutoff,
                                                          cutoff_npeaks[n],
                                                          cutoff_lpeaks[n],
                                                          cutoff_lpeaks[n]/cutoff_npeaks[n]))
        ret = ''.join(ret_list)
        return ret

    @cython.ccall
    def call_peaks(self,
                   cutoff: cython.float = 5.0,
                   min_length: cython.int = 200,
                   max_gap: cython.int = 50,
                   call_summits: bool = False):
        """Return peaks where scores remain above ``cutoff``.

        Args:
            cutoff: Minimum score threshold (e.g., ``-log10 p``).
            min_length: Minimum peak length in bases.
            max_gap: Maximum distance between merged segments.
            call_summits: Whether to report all local maxima within peaks.
        """
        i: cython.int
        chrom: bytes
        pos: cnp.ndarray
        sample: cnp.ndarray
        control: cnp.ndarray
        value: cnp.ndarray
        above_cutoff: cnp.ndarray
        above_cutoff_v: cnp.ndarray
        above_cutoff_endpos: cnp.ndarray
        above_cutoff_startpos: cnp.ndarray
        above_cutoff_sv: cnp.ndarray
        peak_content: list

        chrs = self.get_chr_names()
        peaks = PeakIO()                      # dictionary to save peaks

        self.cutoff = cutoff
        for chrom in sorted(chrs):
            peak_content = []           # to store points above cutoff

            pos = self.data[chrom][0]
            sample = self.data[chrom][1]
            # control = self.data[chrom][2]
            value = self.data[chrom][3]

            # indices where score is above cutoff
            above_cutoff = np.nonzero(value >= cutoff)[0]
            # scores where score is above cutoff
            above_cutoff_v = value[above_cutoff]
            # end positions of regions where score is above cutoff
            above_cutoff_endpos = pos[above_cutoff]
            # start positions of regions where score is above cutoff
            above_cutoff_startpos = pos[above_cutoff-1]
            # sample pileup height where score is above cutoff
            above_cutoff_sv = sample[above_cutoff]
            if above_cutoff_v.size == 0:
                # nothing above cutoff
                continue

            if above_cutoff[0] == 0:
                # first element > cutoff, fix the first point as
                # 0. otherwise it would be the last item in
                # data[chrom]['pos']
                above_cutoff_startpos[0] = 0

            # first bit of region above cutoff
            peak_content.append((above_cutoff_startpos[0],
                                 above_cutoff_endpos[0],
                                 above_cutoff_v[0],
                                 above_cutoff_sv[0],
                                 above_cutoff[0]))
            for i in range(1, above_cutoff_startpos.size):
                if above_cutoff_startpos[i] - peak_content[-1][1] <= max_gap:
                    # append
                    peak_content.append((above_cutoff_startpos[i],
                                         above_cutoff_endpos[i],
                                         above_cutoff_v[i],
                                         above_cutoff_sv[i],
                                         above_cutoff[i]))
                else:
                    # close
                    if call_summits:
                        self.__close_peak2(peak_content,
                                           peaks,
                                           min_length,
                                           chrom,
                                           max_gap//2)
                    else:
                        self.__close_peak(peak_content,
                                          peaks,
                                          min_length,
                                          chrom)
                    peak_content = [(above_cutoff_startpos[i],
                                     above_cutoff_endpos[i],
                                     above_cutoff_v[i],
                                     above_cutoff_sv[i],
                                     above_cutoff[i]),]

            # save the last peak
            if not peak_content:
                continue
            else:
                if call_summits:
                    self.__close_peak2(peak_content,
                                       peaks,
                                       min_length,
                                       chrom,
                                       max_gap//2)
                else:
                    self.__close_peak(peak_content,
                                      peaks,
                                      min_length,
                                      chrom)

        return peaks

    @cython.cfunc
    def __close_peak(self,
                     peak_content: list,
                     peaks: object,
                     min_length: cython.int,
                     chrom: bytes) -> bool:
        """Close the peak region, output peak boundaries, peak summit
        and scores, then add the peak to peakIO object.

        In this function, we define the peak summit as the middle
        point of the region with the highest score, in this peak. For
        example, if the region of the highest score is from 100 to
        200, the summit is 150. If there are several regions of the
        same 'highest score', we will first calculate the possible
        summit for each such region, then pick a position close to the
        middle index (= (len(highest_regions) + 1) / 2) of these
        summits. For example, if there are three regions with the same
        highest scores, [100,200], [300,400], [600,700], we will first
        find the possible summits as 150, 350, and 650, and then pick
        the middle index, the 2nd, of the three positions -- 350 as
        the final summit. If there are four regions, we pick the 2nd
        as well.

        peaks: a PeakIO object

        """
        summit_pos: cython.int
        tstart: cython.int
        tend: cython.int
        summit_index: cython.int
        i: cython.int
        midindex: cython.int
        summit_value: cython.float
        tvalue: cython.float
        tsummitvalue: cython.float

        peak_length = peak_content[-1][1] - peak_content[0][0]
        if peak_length >= min_length:  # if the peak is too small, reject it
            tsummit = []
            summit_pos = 0
            summit_value = 0
            for i in range(len(peak_content)):
                (tstart, tend, tvalue, tsummitvalue, tindex) = peak_content[i]
                #for (tstart,tend,tvalue,tsummitvalue, tindex) in peak_content:
                if not summit_value or summit_value < tsummitvalue:
                    tsummit = [(tend + tstart) / 2,]
                    tsummit_index = [tindex,]
                    summit_value = tsummitvalue
                elif summit_value == tsummitvalue:
                    # remember continuous summit values
                    tsummit.append(int((tend + tstart) / 2))
                    tsummit_index.append(tindex)
            # the middle of all highest points in peak region is defined as summit
            midindex = int((len(tsummit) + 1) / 2) - 1
            summit_pos = tsummit[midindex]
            summit_index = tsummit_index[midindex]
            if self.scoring_method == ord('q'):
                qscore = self.data[chrom][3][summit_index]
            else:
                # if q value is not computed, use -1
                qscore = -1

            peaks.add(chrom,
                      peak_content[0][0],
                      peak_content[-1][1],
                      summit=summit_pos,
                      peak_score=self.data[chrom][3][summit_index],
                      # should be the same as summit_value
                      pileup=self.data[chrom][1][summit_index],  
                      pscore=get_pscore(self.data[chrom][1][summit_index],
                                        self.data[chrom][2][summit_index]),
                      fold_change=(self.data[chrom][1][summit_index] +
                                   self.pseudocount) / (self.data[chrom][2][summit_index] +
                                                        self.pseudocount),
                      qscore=qscore,
                      )
            # start a new peak
            return True

    @cython.cfunc
    def __close_peak2(self,
                      peak_content: list,
                      peaks: object,
                      min_length: cython.int,
                      chrom: bytes,
                      smoothlen: cython.int = 51,
                      min_valley: cython.float = 0.9) -> bool:
        """Close the peak region, output peak boundaries, peak summit
        and scores, then add the peak to peakIO object.

        In this function, we use signal processing methods to smooth
        the scores in the peak region, find the maxima and enforce the
        peaky shape, and to define the best maxima as the peak
        summit. The functions used for signal processing is 'maxima'
        (with 2nd order polynomial filter) and 'enfoce_peakyness'
        functions in SignalProcessing.pyx.

        peaks: a PeakIO object

        """
        tstart: cython.int
        tend: cython.int
        tmpindex: cython.int
        summit_index: cython.int
        summit_offset: cython.int
        start: cython.int
        end: cython.int
        i: cython.int
        j: cython.int
        start_boundary: cython.int
        tvalue: cython.float
        peakdata: cnp.ndarray(cython.float, ndim=1)
        peakindices: cnp.ndarray(cython.int, ndim=1)
        summit_offsets: cnp.ndarray(cython.int, ndim=1)

        # Add 10 bp padding to peak region so that we can get true minima
        end = peak_content[-1][1] + 10
        start = peak_content[0][0] - 10
        if start < 0:
            start_boundary = 10 + start
            start = 0
        else:
            start_boundary = 10
        peak_length = end - start
        if end - start < min_length:
            return             # if the region is too small, reject it

        peakdata = np.zeros(end - start, dtype='f4')
        peakindices = np.zeros(end - start, dtype='i4')
        for (tstart, tend, tvalue, tsvalue, tmpindex) in peak_content:
            i = tstart - start + start_boundary
            j = tend - start + start_boundary
            peakdata[i:j] = tsvalue
            peakindices[i:j] = tmpindex
        summit_offsets = maxima(peakdata, smoothlen)
        if summit_offsets.shape[0] == 0:
            # **failsafe** if no summits, fall back on old approach #
            return self.__close_peak(peak_content, peaks, min_length, chrom)
        else:
            # remove maxima that occurred in padding
            i = np.searchsorted(summit_offsets,
                                start_boundary)
            j = np.searchsorted(summit_offsets,
                                peak_length + start_boundary,
                                'right')
            summit_offsets = summit_offsets[i:j]

        summit_offsets = enforce_peakyness(peakdata, summit_offsets)
        if summit_offsets.shape[0] == 0:
            # **failsafe** if no summits, fall back on old approach #
            return self.__close_peak(peak_content, peaks, min_length, chrom)

        summit_indices = peakindices[summit_offsets]
        summit_offsets -= start_boundary

        peak_scores = self.data[chrom][3][summit_indices]
        if not (peak_scores > self.cutoff).all():
            return self.__close_peak(peak_content, peaks, min_length, chrom)
        for summit_offset, summit_index in zip(summit_offsets, summit_indices):
            if self.scoring_method == ord('q'):
                qscore = self.data[chrom][3][summit_index]
            else:
                # if q value is not computed, use -1
                qscore = -1
            peaks.add(chrom,
                      start,
                      end,
                      summit=start + summit_offset,
                      peak_score=self.data[chrom][3][summit_index],
                      # should be the same as summit_value
                      pileup=self.data[chrom][1][summit_index],
                      pscore=get_pscore(self.data[chrom][1][summit_index],
                                        self.data[chrom][2][summit_index]),
                      fold_change=(self.data[chrom][1][summit_index] +
                                   self.pseudocount) / (self.data[chrom][2][summit_index] +
                                                        self.pseudocount),
                      qscore=qscore,
                      )
        # start a new peak
        return True

    @cython.cfunc
    def total(self) -> cython.long:
        """Return the number of regions in this object.

        """
        t: cython.long
        chrom: bytes

        t = 0
        for chrom in sorted(self.data.keys()):
            t += self.datalength[chrom]
        return t

    @cython.ccall
    def call_broadpeaks(self,
                        lvl1_cutoff: cython.float = 5.0,
                        lvl2_cutoff: cython.float = 1.0,
                        min_length: cython.int = 200,
                        lvl1_max_gap: cython.int = 50,
                        lvl2_max_gap: cython.int = 400):
        """Return broad peaks constructed from high- and low-cutoff segments.

        Args:
            lvl1_cutoff: Threshold for core enriched segments.
            lvl2_cutoff: Threshold for linking segments.
            min_length: Minimum peak length to report.
            lvl1_max_gap: Maximum gap when merging level-1 segments.
            lvl2_max_gap: Maximum allowed length for linking segments.
        """
        i: cython.int
        chrom: bytes

        assert lvl1_cutoff > lvl2_cutoff, "level 1 cutoff should be larger than level 2."
        assert lvl1_max_gap < lvl2_max_gap, "level 2 maximum gap should be larger than level 1."
        lvl1_peaks = self.call_peaks(cutoff=lvl1_cutoff,
                                     min_length=min_length,
                                     max_gap=lvl1_max_gap)
        lvl2_peaks = self.call_peaks(cutoff=lvl2_cutoff,
                                     min_length=min_length,
                                     max_gap=lvl2_max_gap)
        chrs = lvl1_peaks.peaks.keys()
        broadpeaks = BroadPeakIO()
        # use lvl2_peaks as linking regions between lvl1_peaks
        for chrom in sorted(chrs):
            lvl1peakschrom = lvl1_peaks.peaks[chrom]
            lvl2peakschrom = lvl2_peaks.peaks[chrom]
            lvl1peakschrom_next = iter(lvl1peakschrom).__next__
            tmppeakset = []             # to temporarily store lvl1 region inside a lvl2 region
            # our assumption is lvl1 regions should be included in lvl2 regions
            try:
                lvl1 = lvl1peakschrom_next()
                for i in range(len(lvl2peakschrom)):
                    # for each lvl2 peak, find all lvl1 peaks inside
                    # I assume lvl1 peaks can be ALL covered by lvl2 peaks.
                    lvl2 = lvl2peakschrom[i]

                    while True:
                        if lvl2["start"] <= lvl1["start"] and lvl1["end"] <= lvl2["end"]:
                            tmppeakset.append(lvl1)
                            lvl1 = lvl1peakschrom_next()
                        else:
                            # make a hierarchical broad peak
                            #print lvl2["start"], lvl2["end"], lvl2["score"]
                            self.__add_broadpeak(broadpeaks,
                                                 chrom,
                                                 lvl2,
                                                 tmppeakset)
                            tmppeakset = []
                            break
            except StopIteration:
                # no more strong (aka lvl1) peaks left
                self.__add_broadpeak(broadpeaks,
                                     chrom,
                                     lvl2,
                                     tmppeakset)
                tmppeakset = []
                # add the rest lvl2 peaks
                for j in range(i+1, len(lvl2peakschrom)):
                    self.__add_broadpeak(broadpeaks,
                                         chrom,
                                         lvl2peakschrom[j],
                                         tmppeakset)

        return broadpeaks

    def __add_broadpeak(self,
                        bpeaks,
                        chrom: bytes,
                        lvl2peak: dict,
                        lvl1peakset: list):
        """Internal function to create broad peak.
        """

        blockNum: cython.int
        thickStart: cython.int
        thickEnd: cython.int
        start: cython.int
        end: cython.int
        blockSizes: bytes
        blockStarts: bytes

        start = lvl2peak["start"]
        end = lvl2peak["end"]

        # the following code will add those broad/lvl2 peaks with no strong/lvl1 peaks inside
        if not lvl1peakset:
            # will complement by adding 1bps start and end to this region
            # may change in the future if gappedPeak format was improved.
            bpeaks.add(chrom,
                       start,
                       end,
                       score=lvl2peak["score"],
                       thickStart=(b"%d" % start),
                       thickEnd=(b"%d" % end),
                       blockNum=2,
                       blockSizes=b"1,1",
                       blockStarts=(b"0,%d" % (end-start-1)),
                       pileup=lvl2peak["pileup"],
                       pscore=lvl2peak["pscore"],
                       fold_change=lvl2peak["fc"],
                       qscore=lvl2peak["qscore"])
            return bpeaks

        thickStart = b"%d" % lvl1peakset[0]["start"]
        thickEnd = b"%d" % lvl1peakset[-1]["end"]
        blockNum = int(len(lvl1peakset))
        blockSizes = b",".join([b"%d" % x["length"] for x in lvl1peakset])
        blockStarts = b",".join([b"%d" % (x["start"]-start) for x in lvl1peakset])

        if lvl2peak["start"] != thickStart:
            # add 1bp mark for the start of lvl2 peak
            thickStart = b"%d" % start
            blockNum += 1
            blockSizes = b"1,"+blockSizes
            blockStarts = b"0,"+blockStarts
        if lvl2peak["end"] != thickEnd:
            # add 1bp mark for the end of lvl2 peak
            thickEnd = b"%d" % end
            blockNum += 1
            blockSizes = blockSizes+b",1"
            blockStarts = blockStarts + b"," + (b"%d" % (end-start-1))

        # add to BroadPeakIO object
        bpeaks.add(chrom,
                   start,
                   end,
                   score=lvl2peak["score"],
                   thickStart=thickStart,
                   thickEnd=thickEnd,
                   blockNum=blockNum,
                   blockSizes=blockSizes,
                   blockStarts=blockStarts,
                   pileup=lvl2peak["pileup"],
                   pscore=lvl2peak["pscore"],
                   fold_change=lvl2peak["fc"],
                   qscore=lvl2peak["qscore"])
        return bpeaks

@cython.cclass
class TwoConditionScores:
    """Class for saving two condition comparison scores.
    """
    # dictionary for data of each chromosome
    data: dict
    # length of data array of each chromosome
    datalength: dict
    # factor to apply to cond1 pileup values
    cond1_factor: cython.float
    # factor to apply to cond2 pileup values
    cond2_factor: cython.float
    # the pseudocount used to calcuate LLR
    pseudocount: cython.float
    cutoff: cython.float
    t1bdg: object
    c1bdg: object
    t2bdg: object
    c2bdg: object
    pvalue_stat1: dict
    pvalue_stat2: dict
    pvalue_stat3: dict

    def __init__(self,
                 t1bdg,
                 c1bdg,
                 t2bdg,
                 c2bdg,
                 cond1_factor: cython.float = 1.0,
                 cond2_factor: cython.float = 1.0,
                 pseudocount: cython.float = 0.01,
                 proportion_background_empirical_distribution: cython.float = 0.99999):
        """Initialise a differential score track from paired bedGraphs.

        Args:
            t1bdg: Treatment bedGraph for condition 1.
            c1bdg: Control bedGraph for condition 1.
            t2bdg: Treatment bedGraph for condition 2.
            c2bdg: Control bedGraph for condition 2.
            cond1_factor: Scaling factor applied to condition 1 tracks.
            cond2_factor: Scaling factor applied to condition 2 tracks.
            pseudocount: Pseudocount added when comparing conditions.
            proportion_background_empirical_distribution: Fraction of bases
                treated as background when building empirical distributions.
        """
        # for each chromosome, there is a l*4 matrix. First column: end
        # position of a region; Second: treatment pileup; third:
        # control pileup ; forth: score (can be p/q-value/likelihood
        # ratio/fold-enrichment/subtraction depending on -c setting)
        self.data = {}
        self.datalength = {}
        self.cond1_factor = cond1_factor
        self.cond2_factor = cond2_factor
        self.pseudocount = pseudocount
        self.pvalue_stat1 = {}
        self.pvalue_stat2 = {}
        self.t1bdg = t1bdg
        self.c1bdg = c1bdg
        self.t2bdg = t2bdg
        self.c2bdg = c2bdg
        # self.empirical_distr_llr = [] # save all values in histogram

    @cython.ccall
    def set_pseudocount(self, pseudocount: cython.float):
        """Update the pseudocount used for differential scoring."""
        self.pseudocount = pseudocount

    @cython.ccall
    def build(self):
        """Compute scores from 3 types of comparisons and store them
        in self.data.

        """
        common_chrs: set
        chrname: bytes
        chrom_max_len: cython.int
        # common chromosome names
        common_chrs = self.get_common_chrs()
        for chrname in common_chrs:
            (cond1_treat_ps, cond1_treat_vs) = self.t1bdg.get_data_by_chr(chrname)
            (cond1_control_ps, cond1_control_vs) = self.c1bdg.get_data_by_chr(chrname)
            (cond2_treat_ps, cond2_treat_vs) = self.t2bdg.get_data_by_chr(chrname)
            (cond2_control_ps, cond2_control_vs) = self.c2bdg.get_data_by_chr(chrname)
            chrom_max_len = len(cond1_treat_ps) + len(cond1_control_ps) + len(cond2_treat_ps) + len(cond2_control_ps)
            self.add_chromosome(chrname, chrom_max_len)
            self.build_chromosome(chrname,
                                  cond1_treat_ps, cond1_control_ps,
                                  cond2_treat_ps, cond2_control_ps,
                                  cond1_treat_vs, cond1_control_vs,
                                  cond2_treat_vs, cond2_control_vs)

    @cython.cfunc
    def build_chromosome(self, chrname,
                         cond1_treat_ps, cond1_control_ps,
                         cond2_treat_ps, cond2_control_ps,
                         cond1_treat_vs, cond1_control_vs,
                         cond2_treat_vs, cond2_control_vs):
        """Internal function to calculate scores for three types of comparisons.

        cond1_treat_ps, cond1_control_ps: position of treat and control of condition 1
        cond2_treat_ps, cond2_control_ps: position of treat and control of condition 2
        cond1_treat_vs, cond1_control_vs: value of treat and control of condition 1
        cond2_treat_vs, cond2_control_vs: value of treat and control of condition 2

        """
        c1tp: cython.int
        c1cp: cython.int
        c2tp: cython.int
        c2cp: cython.int
        minp: cython.int
        pre_p: cython.int
        c1tv: cython.float
        c1cv: cython.float
        c2tv: cython.float
        c2cv: cython.float

        c1tpn = iter(cond1_treat_ps).__next__
        c1cpn = iter(cond1_control_ps).__next__
        c2tpn = iter(cond2_treat_ps).__next__
        c2cpn = iter(cond2_control_ps).__next__
        c1tvn = iter(cond1_treat_vs).__next__
        c1cvn = iter(cond1_control_vs).__next__
        c2tvn = iter(cond2_treat_vs).__next__
        c2cvn = iter(cond2_control_vs).__next__

        pre_p = 0

        try:
            c1tp = c1tpn()
            c1tv = c1tvn()

            c1cp = c1cpn()
            c1cv = c1cvn()

            c2tp = c2tpn()
            c2tv = c2tvn()

            c2cp = c2cpn()
            c2cv = c2cvn()

            while True:
                minp = min(c1tp, c1cp, c2tp, c2cp)
                self.add(chrname, pre_p, c1tv, c1cv, c2tv, c2cv)
                pre_p = minp
                if c1tp == minp:
                    c1tp = c1tpn()
                    c1tv = c1tvn()
                if c1cp == minp:
                    c1cp = c1cpn()
                    c1cv = c1cvn()
                if c2tp == minp:
                    c2tp = c2tpn()
                    c2tv = c2tvn()
                if c2cp == minp:
                    c2cp = c2cpn()
                    c2cv = c2cvn()
        except StopIteration:
            # meet the end of either bedGraphTrackI, simply exit
            pass
        return

    @cython.cfunc
    def get_common_chrs(self) -> set:
        """Return chromosome names shared across all input bedGraphs."""
        t1chrs: set
        c1chrs: set
        t2chrs: set
        c2chrs: set
        common: set
        t1chrs = self.t1bdg.get_chr_names()
        c1chrs = self.c1bdg.get_chr_names()
        t2chrs = self.t2bdg.get_chr_names()
        c2chrs = self.c2bdg.get_chr_names()
        common = reduce(lambda x, y: x.intersection(y),
                        (t1chrs, c1chrs, t2chrs, c2chrs))
        return common

    @cython.cfunc
    def add_chromosome(self,
                       chrom: bytes,
                       chrom_max_len: cython.int):
        """Allocate storage for ``chrom`` with capacity ``chrom_max_len``."""
        if chrom not in self.data:
            self.data[chrom] = [np.zeros(chrom_max_len, dtype="i4"),  # pos
                                np.zeros(chrom_max_len, dtype="f4"),  # LLR t1 vs c1
                                np.zeros(chrom_max_len, dtype="f4"),  # LLR t2 vs c2
                                np.zeros(chrom_max_len, dtype="f4")]  # LLR t1 vs t2
            self.datalength[chrom] = 0

    @cython.cfunc
    def add(self,
            chromosome: bytes,
            endpos: cython.int,
            t1: cython.float,
            c1: cython.float,
            t2: cython.float,
            c2: cython.float):
        """Take chr-endpos-sample1-control1-sample2-control2 and
        compute logLR for t1 vs c1, t2 vs c2, and t1 vs t2, then save
        values.

        chromosome: chromosome name in string
        endpos    : end position of each interval in integer
        t1        : Sample 1 ChIP pileup value of each interval in float
        c1        : Sample 1 Control pileup value of each interval in float
        t2        : Sample 2 ChIP pileup value of each interval in float
        c2        : Sample 2 Control pileup value of each interval in float

        *Warning* Need to add regions continuously.
        """
        i: cython.int
        c: list

        i = self.datalength[chromosome]
        c = self.data[chromosome]
        c[0][i] = endpos
        c[1][i] = logLR_asym((t1+self.pseudocount) * self.cond1_factor,
                             (c1+self.pseudocount) * self.cond1_factor)
        c[2][i] = logLR_asym((t2+self.pseudocount) * self.cond2_factor,
                             (c2+self.pseudocount) * self.cond2_factor)
        c[3][i] = logLR_sym((t1+self.pseudocount) * self.cond1_factor,
                            (t2+self.pseudocount) * self.cond2_factor)
        self.datalength[chromosome] += 1
        return

    @cython.ccall
    def finalize(self):
        """
        Adjust array size of each chromosome.

        """
        chrom: bytes
        ln: cython.int
        d: list

        for chrom in sorted(self.data.keys()):
            d = self.data[chrom]
            ln = self.datalength[chrom]
            d[0].resize(ln, refcheck=False)
            d[1].resize(ln, refcheck=False)
            d[2].resize(ln, refcheck=False)
            d[3].resize(ln, refcheck=False)
        return

    @cython.ccall
    def get_data_by_chr(self,
                        chromosome: bytes):
        """Return array of counts by chromosome.

        The return value is a tuple:
        ([end pos],[value])
        """
        if chromosome in self.data:
            return self.data[chromosome]
        else:
            return None

    @cython.ccall
    def get_chr_names(self):
        """Return all the chromosome names stored.

        """
        return set(self.data.keys())

    @cython.ccall
    def write_bedGraph(self,
                       fhd,
                       name: str,
                       description: str,
                       column: cython.int = 3):
        """Write all data to fhd in bedGraph Format.

        fhd: a filehandler to save bedGraph.

        name/description: the name and description in track line.

        colname: can be 1: cond1 chip vs cond1 ctrl, 2: cond2 chip vs
        cond2 ctrl, 3: cond1 chip vs cond2 chip

        """
        chrom: bytes
        ln: cython.int
        pre: cython.int
        i: cython.int
        p: cython.int
        pre_v: cython.float
        v: cython.float
        pos: cnp.ndarray
        value: cnp.ndarray

        assert column in range(1, 4), "column should be between 1, 2 or 3."

        write = fhd.write

        # if self.trackline:
        #    # this line is REQUIRED by the wiggle format for UCSC browser
        #    write("track type=bedGraph name=\"%s\" description=\"%s\"\n" % (name.decode(), description))

        chrs = self.get_chr_names()
        for chrom in sorted(chrs):
            pos = self.data[chrom][0]
            value = self.data[chrom][column]
            ln = self.datalength[chrom]
            pre = 0
            if pos.shape[0] == 0:
                continue        # skip if there's no data
            pre_v = value[0]
            for i in range(1, ln):
                v = value[i]
                p = pos[i-1]
                if abs(pre_v - v) >= 1e-6:
                    write("%s\t%d\t%d\t%.5f\n" %
                          (chrom.decode(), pre, p, pre_v))
                    pre_v = v
                    pre = p
            p = pos[-1]
            # last one
            write("%s\t%d\t%d\t%.5f\n" % (chrom.decode(), pre, p, pre_v))

        return True

    @cython.ccall
    def write_matrix(self,
                     fhd,
                     name: str,
                     description: str):
        """Write all data to fhd into five columns Format:

        col1: chr_start_end
        col2: t1 vs c1
        col3: t2 vs c2
        col4: t1 vs t2

        fhd: a filehandler to save the matrix.

        """
        chrom: bytes
        ln: cython.int
        pre: cython.int
        i: cython.int
        p: cython.int
        v1: cython.float
        v2: cython.float
        v3: cython.float
        pos: cnp.ndarray
        value1: cnp.ndarray
        value2: cnp.ndarray
        value3: cnp.ndarray

        write = fhd.write

        chrs = self.get_chr_names()
        for chrom in sorted(chrs):
            [pos, value1, value2, value3] = self.data[chrom]
            ln = self.datalength[chrom]
            pre = 0
            if pos.shape[0] == 0:
                continue        # skip if there's no data
            for i in range(0, ln):
                v1 = value1[i]
                v2 = value2[i]
                v3 = value3[i]
                p = pos[i]
                write("%s:%d_%d\t%.5f\t%.5f\t%.5f\n" %
                      (chrom.decode(), pre, p, v1, v2, v3))
                pre = p

        return True

    @cython.ccall
    def call_peaks(self,
                   cutoff: cython.float = 3,
                   min_length: cython.int = 200,
                   max_gap: cython.int = 100,
                   call_summits: bool = False) -> tuple:
        """This function try to find regions within which, scores
        are continuously higher than a given cutoff.

        For bdgdiff.

        This function is NOT using sliding-windows. Instead, any
        regions in bedGraph above certain cutoff will be detected,
        then merged if the gap between nearby two regions are below
        max_gap. After this, peak is reported if its length is above
        min_length.

        cutoff:  cutoff of value, default 3. For log10 LR, it means 1000 or -1000.
        min_length :  minimum peak length, default 200.
        max_gap   :  maximum gap to merge nearby peaks, default 100.
        ptrack:  an optional track for pileup heights. If it's not None, use it to find summits. Otherwise, use self/scoreTrack.
        """
        chrom: bytes
        pos: cnp.ndarray
        t1_vs_c1: cnp.ndarray
        t2_vs_c2: cnp.ndarray
        t1_vs_t2: cnp.ndarray
        cond1_over_cond2: cnp.ndarray
        cond2_over_cond1: cnp.ndarray
        cond1_equal_cond2: cnp.ndarray
        cond1_sig: cnp.ndarray
        cond2_sig: cnp.ndarray
        cat1: cnp.ndarray
        cat2: cnp.ndarray
        cat3: cnp.ndarray
        cat1_startpos: cnp.ndarray
        cat1_endpos: cnp.ndarray
        cat2_startpos: cnp.ndarray
        cat2_endpos: cnp.ndarray
        cat3_startpos: cnp.ndarray
        cat3_endpos: cnp.ndarray

        chrs = self.get_chr_names()
        cat1_peaks = PeakIO()       # dictionary to save peaks significant at condition 1
        cat2_peaks = PeakIO()       # dictionary to save peaks significant at condition 2
        cat3_peaks = PeakIO()       # dictionary to save peaks significant in both conditions

        self.cutoff = cutoff

        for chrom in sorted(chrs):
            pos = self.data[chrom][0]
            t1_vs_c1 = self.data[chrom][1]
            t2_vs_c2 = self.data[chrom][2]
            t1_vs_t2 = self.data[chrom][3]
            and_ = np.logical_and
            # regions with stronger cond1 signals
            cond1_over_cond2 = t1_vs_t2 >= cutoff
            # regions with stronger cond2 signals
            cond2_over_cond1 = t1_vs_t2 <= -1*cutoff
            cond1_equal_cond2 = and_(t1_vs_t2 >= -1*cutoff, t1_vs_t2 <= cutoff)
            # enriched regions in condition 1
            cond1_sig = t1_vs_c1 >= cutoff
            # enriched regions in condition 2
            cond2_sig = t2_vs_c2 >= cutoff
            # indices where score is above cutoff
            # cond1 stronger than cond2, the indices
            cat1 = np.where(and_(cond1_sig, cond1_over_cond2))[0]
            # cond2 stronger than cond1, the indices
            cat2 = np.where(and_(cond2_over_cond1, cond2_sig))[0]
            # cond1 and cond2 are equal, the indices
            cat3 = np.where(and_(and_(cond1_sig, cond2_sig),
                                 cond1_equal_cond2))[0]

            # end positions of regions where score is above cutoff
            cat1_endpos = pos[cat1]
            # start positions of regions where score is above cutoff            
            cat1_startpos = pos[cat1-1]
            # end positions of regions where score is above cutoff
            cat2_endpos = pos[cat2]
            # start positions of regions where score is above cutoff
            cat2_startpos = pos[cat2-1]
            # end positions of regions where score is above cutoff
            cat3_endpos = pos[cat3]
            # start positions of regions where score is above cutoff
            cat3_startpos = pos[cat3-1]

            # for cat1: condition 1 stronger regions
            self.__add_a_peak(cat1_peaks,
                              chrom,
                              cat1,
                              cat1_startpos,
                              cat1_endpos,
                              t1_vs_t2,
                              max_gap,
                              min_length)
            # for cat2: condition 2 stronger regions
            self.__add_a_peak(cat2_peaks,
                              chrom,
                              cat2,
                              cat2_startpos,
                              cat2_endpos,
                              -1 * t1_vs_t2,
                              max_gap,
                              min_length)
            # for cat3: commonly strong regions
            self.__add_a_peak(cat3_peaks,
                              chrom,
                              cat3,
                              cat3_startpos,
                              cat3_endpos,
                              abs(t1_vs_t2),
                              max_gap,
                              min_length)

        return (cat1_peaks, cat2_peaks, cat3_peaks)

    @cython.cfunc
    def __add_a_peak(self,
                     peaks: object,
                     chrom: bytes,
                     indices: cnp.ndarray,
                     startpos: cnp.ndarray,
                     endpos: cnp.ndarray,
                     score: cnp.ndarray,
                     max_gap: cython.int,
                     min_length: cython.int):
        
        """For a given chromosome, merge nearby significant regions,
        filter out smaller regions, then add regions to PeakIO
        object.

        """
        i: cython.int
        peak_content: list
        mean_logLR: cython.float

        if startpos.size > 0:
            # if it is not empty
            peak_content = []
            if indices[0] == 0:
                # first element > cutoff, fix the first point as
                # 0. otherwise it would be the last item in
                # data[chrom]['pos']
                startpos[0] = 0
            # first bit of region above cutoff
            peak_content.append((startpos[0],
                                 endpos[0],
                                 score[indices[0]]))
            for i in range(1, startpos.size):
                if startpos[i] - peak_content[-1][1] <= max_gap:
                    # append
                    peak_content.append((startpos[i],
                                         endpos[i],
                                         score[indices[i]]))
                else:
                    # close
                    if peak_content[-1][1] - peak_content[0][0] >= min_length:
                        mean_logLR = self.mean_from_peakcontent(peak_content)
                        # if peak_content[0][0] == 22414956:
                        #    print(f"{peak_content} {mean_logLR}")
                        peaks.add(chrom,
                                  peak_content[0][0],
                                  peak_content[-1][1],
                                  summit=-1,
                                  peak_score=mean_logLR,
                                  pileup=0,
                                  pscore=0,
                                  fold_change=0,
                                  qscore=0,
                                  )
                    peak_content = [(startpos[i],
                                     endpos[i],
                                     score[indices[i]]),]

            # save the last peak
            if peak_content:
                if peak_content[-1][1] - peak_content[0][0] >= min_length:
                    mean_logLR = self.mean_from_peakcontent(peak_content)
                    peaks.add(chrom,
                              peak_content[0][0],
                              peak_content[-1][1],
                              summit=-1,
                              peak_score=mean_logLR,
                              pileup=0,
                              pscore=0,
                              fold_change=0,
                              qscore=0,
                              )

        return

    @cython.cfunc
    def mean_from_peakcontent(self,
                              peakcontent: list) -> cython.float:
        """

        """
        tmp_s: cython.int
        tmp_e: cython.int
        ln: cython.int
        tmp_v: cython.long
        sum_v: cython.long      # for better precision
        r: cython.float
        i: cython.int

        ln = 0
        sum_v = 0                         # initialize sum_v as 0
        for i in range(len(peakcontent)):
            tmp_s = peakcontent[i][0]
            tmp_e = peakcontent[i][1]
            tmp_v = peakcontent[i][2]
            sum_v += tmp_v * (tmp_e - tmp_s)
            ln += tmp_e - tmp_s

        r = cython.cast(cython.float, (sum_v / ln))
        return r

    @cython.cfunc
    def total(self) -> cython.long:
        """Return the number of regions in this object.

        """
        t: cython.long
        chrom: bytes

        t = 0
        for chrom in sorted(self.data.keys()):
            t += self.datalength[chrom]
        return t



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2025-02-05 12:38:24 Tao Liu>

"""Utilities for reading and writing MACS3 bedGraph files.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
from array import array

from MACS3.Signal.BedGraph import bedGraphTrackI
import cython
from cython.cimports.cpython import bool

# ------------------------------------
# constants
# ------------------------------------

# ------------------------------------
# C lib
# ------------------------------------
from cython.cimports.libc.stdlib import atoi, atof

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Classes
# ------------------------------------


@cython.cclass
class bedGraphIO:
    """Helper for loading and writing bedGraph tracks."""
    bedGraph_filename = cython.declare(str, visibility='public')
    data = cython.declare(object, visibility='public')

    def __init__(self, bedGraph_filename: str, data=None):
        """Initialise the IO wrapper for ``bedGraph_filename``.

        Args:
            bedGraph_filename: Path to the bedGraph file.
            data: Optional existing :class:`bedGraphTrackI` to populate.
        """
        self.bedGraph_filename = bedGraph_filename
        if data:
            assert isinstance(data, bedGraphTrackI)
            self.data = data
        else:
            self.data = bedGraphTrackI()

    @cython.ccall
    def read_bedGraph(self, baseline_value: cython.double = 0):
        """Load bedGraph intervals into the internal :class:`bedGraphTrackI`.

        Args:
            baseline_value: Value used to fill gaps between defined intervals.

        Returns:
            bedGraphTrackI: Populated track instance.

        Example:
            .. code-block:: python
            
                from MACS3.IO.BedGraphIO import bedGraphIO
                bio = bedGraphIO("signal.bdg")
                track = bio.read_bedGraph(baseline_value=0)
        """
        i: bytes

        self.data.reset_baseline(baseline_value)
        add_func = self.data.add_loc
        # python open file
        bedGraph_file = open(self.bedGraph_filename, "rb")

        for i in bedGraph_file:
            if i.startswith(b"track"):
                continue
            elif i.startswith(b"#"):
                continue
            elif i.startswith(b"browse"):
                continue
            else:
                fs = i.split()
                add_func(fs[0], atoi(fs[1]), atoi(fs[2]), atof(fs[3]))

        bedGraph_file.close()
        return self.data

    @cython.ccall
    def write_bedGraph(self, name: str = "", description: str = "",
                       trackline: bool = True):
        """Persist the current track to ``self.bedGraph_filename``.

        Args:
            name: Track name used in the optional header line.
            description: Track description used in the optional header line.
            trackline: Whether to emit a UCSC ``track`` header.

        Examples:
            .. code-block:: python

                from MACS3.IO.BedGraphIO import bedGraphIO
                bio = bedGraphIO("signal_out.bdg")
                bio.write_bedGraph(name="Example", description="Demo track")
        """
        pre: cython.int
        pos: cython.int
        i: cython.int
        value: cython.double
        chrom: bytes
        chrs: set
        trackcontents: tuple
        p: array
        v: array

        fhd = open(self.bedGraph_filename, "w")
        if trackline:
            trackcontents = (name.replace("\"", "\\\""),
                             description.replace("\"", "\\\""))
            fhd.write("track type=bedGraph name=\"%s\" description=\"%s\" visibility=2 alwaysZero=on\n" % trackcontents)
        chrs = self.data.get_chr_names()
        for chrom in sorted(chrs):
            (p, v) = self.data.get_data_by_chr(chrom)
            pnext = iter(p).__next__
            vnext = iter(v).__next__
            pre = 0

            for i in range(len(p)):
                pos = pnext()
                value = vnext()
                fhd.write("%s\t%d\t%d\t%.5f\n" %
                          (chrom.decode(), pre, pos, value))
                pre = pos
        fhd.close()
        return
